package sita;

import java.awt.Point;
import java.util.ArrayList;

/*
 * Represents the information in one annotation.
 * An annotation can consist of three parts
 *      1. location of left, right brackets or box on manuscript image
 *      2. location of highlighted digital text
 *      3. comment
*/
public class AnnotationInfo
{
    private String encoding;
    private int tabNumber = -1;
    private int sequenceNumber;                                 // to keep track of original order when sorting, not saved in xml file.
    private Point p1 = new Point(-1, -1);               // the coordinate of either the left bracket or the upper left corner of the box
    private Point normalizedP1 = new Point(-1, -1);     // version of p1 that has been adjusted to the camera image size. This is what is stored in the XML file.

    private Point p2 = new Point(-1, -1);               // the coordinate of either the right bracket or the lower right corner of the box
    private Point normalizedP2 = new Point(-1, -1);     // version of p1 that has been adjusted to the camera image size. This is what is stored in the XML file.

    private Point savep1;
    private Point saveNormalizedP1;
    private PrecisePoint savepp1;

    private ArrayList<TextBlock>textMarkers = new ArrayList<TextBlock>();
    private String comment = "";
    private boolean leftBracketExists = false;
    private boolean rightBracketExists = false;
    private boolean boxExists = false;
    private boolean needsNormalization = false;

    private PrecisePoint pp1;
    private PrecisePoint pp2;

    public AnnotationInfo()
    {
    }


    // Initializes as an empty annotation
    public void clear()
    {
        encoding = null;
        tabNumber = -1;
        p1.x = -1;
        p1.y = -1;
        p2.x = -1;
        p2.y = -1;
        pp1 = null;
        pp2 = null;
        textMarkers.clear();
        comment = "";
        leftBracketExists = false;
        leftBracketExists = false;
        boxExists = false;
        needsNormalization = false;
    }

    /*************************************************************
     *  The following set of methods support the marking of the manuscript image with either brackets or a box.
     *************************************************************/

    // Save location of left bracket (or upper left corner of box).
    // This is in current screen coordinates and needs to be normalized before it is saved in XML file
    public void setPoint1(Point p)
    {
        p1.x = p.x;
        p1.y = p.y;
        needsNormalization = true;
    }

    // Variation on previous setPoint1
    public void setPoint1(int x, int y)
    {
        p1.x = x;
        p1.y = y;
        needsNormalization = true;
    }

    // Returns location of left bracket (or upper left corner of box)
    public Point getPoint1()
    {
        return p1;
    }


    // Save location of right bracket (or lower right corner of box).
    // This is in current screen coordinates and needs to be normalized before it is saved in XML file
    public void setPoint2(Point p)
    {
        p2.x = p.x;
        p2.y = p.y;
        needsNormalization = true;
    }


    // Variation on previous setPoint2
    public void setPoint2(int x, int y)
    {
        p2.x = x;
        p2.y = y;
        needsNormalization = true;
    }


    // Returns location of right bracket (or lower right corner of box)
    public Point getPoint2()
    {
        return p2;
    }

    /*
     * The following 12 methods have form
     *      setNormalized_, getNormalized_, get_ where the blank is either X1, Y1, X2, Y2
     *      They provide an alternate way of modifying p1, normalizedP1, p2, normalizedP2
     */
    public void setNormalizedX1(int x1)
    {
        normalizedP1.x = x1;
        leftBracketExists = normalizedP1.x != -1 && normalizedP1.y != -1;
        boxExists = leftBracketExists && rightBracketExists;
    }

    public int getNormalizedX1()
    {
        return normalizedP1.x;
    }

    public int getX1()
    {
        return p1.x;
    }


    public void setNormalizedY1(int y1)
    {
        normalizedP1.y = y1;
        leftBracketExists = normalizedP1.x != -1 && normalizedP1.y != -1;
        boxExists = leftBracketExists && rightBracketExists;
    }

    public int getNormalizedY1()
    {
        return normalizedP1.y;
    }

    public int getY1()
    {
        return p1.y;
    }


    public void setNormalizedX2(int x2)
    {
        normalizedP2.x = x2;
        rightBracketExists = normalizedP2.x != -1 && normalizedP2.y != -1;
        boxExists = leftBracketExists && rightBracketExists;
    }

    public int getNormalizedX2()
    {
        return normalizedP2.x;
    }

    public int getX2()
    {
        return p2.x;
    }


    public void setNormalizedY2(int y2)
    {
        normalizedP2.y = y2;
        rightBracketExists = normalizedP2.x != -1 && normalizedP2.y != -1;
        boxExists = leftBracketExists && rightBracketExists;
    }

    public int getNormalizedY2()
    {
        return normalizedP2.y;
    }

    public int getY2()
    {
        return p2.y;
    }

    public void setEncoding (String encoding) {
        this.encoding = encoding;
    }

    public String getEncoding () {
        return this.encoding;
    }

    public void setTabNumber (int tabNumber) {
        this.tabNumber = tabNumber;
    }

    public int getTabNumber () {
        return this.tabNumber;
    }

    public boolean swapBrackets()
    {
        boolean rc = false;

        if (leftBracketExists && rightBracketExists)
        {
            Point swap;
            swap = p1;
            p1 = p2;
            p2 = swap;
            
            swap = normalizedP1;
            normalizedP1 = normalizedP2;
            normalizedP2 = swap;
            
            PrecisePoint ppswap;
            ppswap = pp1;
            pp1 = pp2;
            pp2 = ppswap;

            rc = true;
        }

        return rc;
    }

    /*******************************************************************
     * The following set of methods support the highlighting of digital text
     * All values are in characters, not pixels.
     *******************************************************************/

    public void clearTextMarkers () {
        textMarkers.clear();
    }

    // Save offset of start of highlight or -1 if no highlight.
    public void addTextMarker(String filename, TextPoint start, TextPoint end)
    {
        textMarkers.add(new TextBlock(filename, start.row, start.col, end.row, (int)end.col));
    }

    public void addTextMarker(String filename, int startRow, int startCol, int endRow, int endCol)
    {
        textMarkers.add(new TextBlock(filename, startRow, startCol, endRow, endCol));
    }

    public int getSequenceNumber()
    {
        return sequenceNumber;
    }

    public void setSequenceNumber(int sequenceNumber)
    {
        this.sequenceNumber = sequenceNumber;
    }


    /*************************************************************************
     * The following methods support the comment portion of the annotation
     *************************************************************************/
    public void setComment(String comment)
    {
        this.comment = comment;
    }

    public String getComment()      // !! Is there a better way of doing this? Need to get comment from text box and don't want to create new object for each character typed.
    {
        return comment;
    }

    public ArrayList<TextBlock> getTextMarkers () {
        return textMarkers;
    }


    public TextPoint getTextStart (ArrayList<TextMap> textmap) {
        if (textMarkers.size() ==0) {
            return new TextPoint(-1, -1);
        }
        else {
            TextBlock tb = textMarkers.get(0);
            return new TextPoint(mapOffset(textmap, tb, tb.startRow), tb.startCol);
        }
    }

    public TextPoint getTextEnd (ArrayList<TextMap> textmap) {
        if (textMarkers.size() ==0) {
            return new TextPoint(-1, -1);
        }
        else {
            TextBlock tb = textMarkers.get(textMarkers.size() - 1);
            return new TextPoint(mapOffset(textmap, tb, tb.endRow), tb.endCol);
        }
    }

    public int mapOffset (ArrayList<TextMap> textmap, TextBlock tb, int offset) {

        for (TextMap tm : textmap) {
            if (tm.filename != null && tm.filename.equals(tb.filename)) {
                if (tb.startRow >= tm.textStartrow && tb.endRow <= tm.textEndrow) {
                    return tm.tabStartrow + offset - tm.textStartrow;
                }
            }
        }
        return -1;
    }

    public void setLeftBracketExists(boolean leftBracketExists)
    {
        boolean b = this.leftBracketExists;
        this.leftBracketExists = leftBracketExists;

        if (!leftBracketExists)
        {
            p1.x = -1;
            p1.y = -1;
            p2.x = -1;
            p2.y = -1;
            pp1 = null;
            pp2 = null;
        }

        if (b != leftBracketExists)
            this.pp1 = null;    // force this to be reinitialized
    }

    public boolean getLeftBracketExists()
    {
        return leftBracketExists;
    }


    //
    public void setRightBracketExists(boolean rightBracketExists)
    {
        boolean b = this.rightBracketExists;
        this.rightBracketExists = rightBracketExists;

        if (!rightBracketExists)
        {
            p2.x = -1;
            p2.y = -1;
            pp2 = null;
        }
        if (b != rightBracketExists)
            this.pp2 = null;    // force this to be reinitialized
    }

    public boolean getRightBracketExists()
    {
        return rightBracketExists;
    }

    public void setBoxExists(boolean boxExists)
    {
        boolean b = this.boxExists;

        this.boxExists = boxExists;

        if (!boxExists)
            clearBox();

        if (b != boxExists)
        {
            this.pp1 = null;              // force these to be reset.
            this.pp2 = null;
        }
    }

    public void clearBox()
    {
        p1.x = -1;
        p1.y = -1;
        p2.x = -1;
        p2.y = -1;
        pp1 = null;
        pp2 = null;
    }

    public void rememberStartPoint()
    {
        if (p1 !=  null)
            savep1 = new Point(p1);
        else
            savep1 = null;

        if (saveNormalizedP1 != null)
            saveNormalizedP1 = new Point(normalizedP1);
        else
            saveNormalizedP1 = null;

        if (pp1 != null)
            savepp1 = new PrecisePoint(pp1);
        else
            savepp1 =  null;
    }

    public void restoreStartPoint()
    {
        if (savep1 != null)
            p1 = savep1;

        if (saveNormalizedP1 != null)
            normalizedP1 = saveNormalizedP1;

        if (savepp1 != null)
            pp1 = savepp1;
    }

    public boolean getBoxExists()
    {
        return boxExists;
    }




    public void normalize(int before, int after)       // adjust points to be correct in actual image (make them bigger)
    {
        if (needsNormalization)        // if not dirty then don't change original normalized values;
        {
            if (p1.x != -1)
                Utils.adjustPoint(p1, normalizedP1, before, after);

            if (p2.x != -1)
                Utils.adjustPoint(p2, normalizedP2, before, after);

            needsNormalization = false;
        }
    }


    public void denormalize(int before, int after)       // adjust points to be correct in displayed image (make them smaller)
    {
        if (normalizedP1.x != -1)
            Utils.adjustPoint(normalizedP1, p1, before, after);

        if (normalizedP2.x != -1)
            Utils.adjustPoint(normalizedP2, p2, before, after);
    }


    // This is called when user zooms in or out. To reduce roundoff error, each point is first normalized using the pre-zoom image width (oldWidth)
    // and then denormalized using the zoom width (newWidth).
    // Note: this does not change the normalized value of the point!!
    public void adjustPoints(int actualWidth, int oldWidth, int newWidth)
    {
        if (p1.x != -1)
        {
            if (pp1 == null)
                pp1 = new PrecisePoint(p1);
            Utils.adjustPoint(pp1, oldWidth, actualWidth);   // normalize the start point
            Utils.adjustPoint(pp1, actualWidth, newWidth);   // denormalize it to the new width
            pp1.setPoint(p1);
        }

        if (p2.x != -1)
        {
            if (pp2 == null)
                pp2 = new PrecisePoint(p2);
            Utils.adjustPoint(pp2, oldWidth, actualWidth);   // normalize the end point
            Utils.adjustPoint(pp2, actualWidth, newWidth);   // denormalize it to the new width
            pp2.setPoint(p2);
        }
    }

    
    public boolean hasManuscriptAnnotation()
    {
        return leftBracketExists || boxExists;
    }

    public boolean hasSelectedText()
    {
        return textMarkers.size() > 0;
    }

    public boolean isEmpty()
    {
        return !hasManuscriptAnnotation() && !hasSelectedText() && (comment == null || comment.length() == 0);
    }

    public boolean compare(AnnotationInfo ai)
    {
        int len1 = textMarkers.size();
        int len2 = ai.textMarkers.size();
        if (len1 != len2) {
            return false;
        }
        for (int i=0; i<len1; i++) {
            TextBlock tb1 = textMarkers.get(i);
            TextBlock tb2 = ai.textMarkers.get(i);
            if (tb1.startRow == tb2.startRow && tb1.startCol == tb2.startCol && tb1.endRow == tb2.endRow && tb1.endCol == tb2.endCol) {
                continue;
            } else {
                return false;
            }
        }
        
        return
               normalizedP1.x == ai.normalizedP1.x &&
               normalizedP1.y == ai.normalizedP1.y &&
               normalizedP2.x == ai.normalizedP2.x &&
               normalizedP2.y == ai.normalizedP2.y &&
               comment.equals(ai.comment);
    }

    @Override
    public String toString () {
        StringBuilder sb = new StringBuilder();

        return sb.toString();
    }
}
