package sita;

import java.io.File;
import java.io.BufferedWriter;
import java.util.ArrayList;
import java.awt.Color;
import java.io.FileWriter;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class Annotations extends DefaultHandler //implements ErrorHandler, ContentHandler, DTDHandler, EntityResolver
{
    static private int defaultPointSize = 16;
    static private Color defaultPointColor = Color.red;
    private Locator locator;
    private ArrayList<String> elements;
    private ArrayList<AnnotationInfo> annotations;
    private AnnotationInfo currentAnnotation;
    private String dirname;
    private int originalResolvedWidth = 0;
    private int resolvedWidth = 0;
    private int pointSize = defaultPointSize;
    private Color pointColor = defaultPointColor;
    private ApplicationState appstate;
    StringBuilder comment = new StringBuilder();
    String annotationFilename = null;

    public Annotations(ApplicationState appstate)
    {
        this.appstate = appstate;
        elements = new ArrayList<String>();
        annotations = new ArrayList<AnnotationInfo>();
    }


    public void loadAnnotations(String rootdir, String dirname, String pagename) throws Exception
    {
        this.dirname = dirname;
        annotationFilename = String.format("%s/%s/%s", rootdir, dirname, pagename);
        File f = new File(annotationFilename);
        if (!f.exists())
        {
            annotations.clear();
            appstate.setCurrentAnnotationInfo(null);
            f = new File(String.format("%s/%s", rootdir, dirname));
            f.mkdirs();      // in case directory does not exist
            BufferedWriter bw = IOAssistance.openWriter(annotationFilename);
            resolvedWidth = appstate.getDefaultWidth();
            pointColor = appstate.getDefaultPointColor();
            bw.write(String.format("<annotations dir='%s' width='%d' pc='%d'>\n</annotations>",
                    dirname, resolvedWidth, pointColor.getRGB()));
            bw.close();
        }
        else
            loadAnnotations(dirname, annotationFilename);
    }


    public void loadAnnotations(String dirname, String filename) throws Exception
    {
        elements.clear();
        annotations.clear();
        this.dirname = dirname;

        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        saxParser.parse(new File(filename), this);
    }


    public void removeAnnotation(AnnotationInfo ai)
    {
        annotations.remove(ai);
    }

    public void setResolvedWidth(int resolvedWidth)
    {
        this.resolvedWidth = resolvedWidth;
    }

    public int getResolvedWidth()
    {
        return resolvedWidth;
    }


    public void setPointSize(int pointSize)
    {
        this.pointSize = pointSize;
    }

    public int getPointSize()
    {
        return pointSize;
    }

    public void setPointColor(Color pointColor)
    {
        this.pointColor = pointColor;
    }

    public Color getPointColor()
    {
        return pointColor;
    }

    public String getDirname()
    {
        return dirname;
    }

    public ArrayList<AnnotationInfo> getAnnotations()
    {
        return annotations;
    }

    public AnnotationInfo getAnnotation(int n)
    {
        return annotations.get(n);
    }

    public boolean contains(AnnotationInfo ai)
    {
        return (annotations != null && annotations.contains(ai));
    }

    public void delete(AnnotationInfo ai)
    {
        annotations.remove(ai);
    }

    public void delete(int idx)
    {
        annotations.remove(idx);
    }

    // For ContentHandler
    @Override
    public void setDocumentLocator(Locator locator)
    {
        this.locator = locator;
    }

    @Override

    public void startDocument() throws SAXException
    {
    }

    @Override
    public void endDocument() throws SAXException
    {
    }


    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        elements.add(qName);
        if (qName.equals("annotations")) {
            String value = atts.getValue("width");
            if (value != null) {
                resolvedWidth = Integer.parseInt(value);
                originalResolvedWidth = resolvedWidth;
            }
            // otherwise use previous resolved width


            value = atts.getValue("pc");
            if (value != null) {
                pointColor = new Color(Integer.parseInt(value));
            }
            // otherwise use previous point color
        } else if (qName.equals("annotation")) {
            currentAnnotation = new AnnotationInfo();
            String encoding = atts.getValue("encoding");
            if (encoding == null) {
                encoding = "?";
            }
            currentAnnotation.setEncoding(encoding);

            int tabNumber = -1;
            if (atts.getValue("tab") != null) {
                tabNumber = Integer.parseInt(atts.getValue("tab"));
            }
            currentAnnotation.setTabNumber(tabNumber);
        }
        else if (qName.equals("manuscript")) {
            String value = atts.getValue("x1");
            if (value != null) {
                currentAnnotation.setNormalizedX1(Integer.parseInt(value));
            }

            value = atts.getValue("y1");
            if (value != null) {
                currentAnnotation.setNormalizedY1(Integer.parseInt(value));
            }

            value = atts.getValue("x2");
            if (value != null) {
                currentAnnotation.setNormalizedX2(Integer.parseInt(value));
            }

            value = atts.getValue("y2");
            if (value != null) {
                currentAnnotation.setNormalizedY2(Integer.parseInt(value));
            }
        } else if (qName.equals("text")) {
            String filename = atts.getValue("filename");
            String startRow = atts.getValue("startRow");
            String startCol = atts.getValue("startCol");
            String endRow = atts.getValue("endRow");
            String endCol = atts.getValue("endCol");
            if (startRow != null && startCol != null && endRow != null && endCol != null) {
                try {
                    currentAnnotation.addTextMarker(filename, Integer.parseInt(startRow), Integer.parseInt(startCol),
                                                    Integer.parseInt(endRow), Integer.parseInt(endCol));
                } catch (Exception ex) {
                    System.err.format("%s/%s.xml non-integer field in text marker\n", dirname, filename);
                }
            } else {
                System.err.format("%s/%s.xml incomplete text marker\n", dirname, filename);
            }
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        elements.remove(elements.size()-1);
        if (qName.equals("annotation"))
        {
            currentAnnotation.setComment(comment.toString());
            comment.setLength(0);
            annotations.add(currentAnnotation);
            currentAnnotation = null;
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        String currentElement = elements.get(elements.size()-1);
        if (currentElement.equals("comment"))
        {
            String content = new String(ch, start, length);
            comment.append(content);
        }
    }


    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        characters(ch, start, length);
    }
    

    public void msg(String text) throws SAXException
    {
        throw new SAXException(String.format("%d.%d %s", locator.getLineNumber(), locator.getColumnNumber(), text));
    }


    public void add(AnnotationInfo ai)
    {
        annotations.add(ai);
    }


    public int getSize()
    {
        int size = 0;

        if (annotations != null)
            size = annotations.size();

        return size;
    }


    public void clear () {
        annotations.clear();
    }

    public void toXml(int resolvedWidth) throws Exception
    {
        try {
            File f = new File(annotationFilename);
            long lastModified = f.lastModified();
            FileWriter fw = IOAssistance.openUncachedWriter(annotationFilename);
            fw.write(toString());
            fw.flush();
            fw.close();
        } catch (Exception ex) {
            throw new Exception(String.format("Changes not saved! Error was %s", ex.getMessage()));
        }
    }

    @Override
    public String toString () {
        if (appstate.getUseDefaultWidth())                  // if we are ignoring width that user set and using the default width, ignore any changes that we might have made.
            resolvedWidth = originalResolvedWidth;

        StringBuilder sb = new StringBuilder();

        sb.append(String.format("<annotations version='2.0' dir='%s' width='%d' pc='%d'>\n",
                dirname, resolvedWidth, pointColor.getRGB()));
        for (AnnotationInfo ai : annotations)
        {
            if (ai.isEmpty())
                continue;

            sb.append(String.format("\t<annotation encoding='%s' tab='%d'>\n", ai.getEncoding(), ai.getTabNumber()));

            // write manuscript annotation, if any
            if (ai.getLeftBracketExists() || ai.getRightBracketExists())
            {
                sb.append("\t\t<manuscript ");
                if (ai.getLeftBracketExists()) {
                    sb.append(String.format("x1='%d' ", ai.getNormalizedX1()));
                    sb.append(String.format("y1='%d' ", ai.getNormalizedY1()));
                }

                if (ai.getRightBracketExists()) {
                    sb.append(String.format("x2='%d' ", ai.getNormalizedX2()));
                    sb.append(String.format("y2='%d' ", ai.getNormalizedY2()));
                }

                sb.append("/>\n");
            }

            // write text markers, if any
            ArrayList<TextBlock> textMarkers = ai.getTextMarkers();
            for (TextBlock textMarker : textMarkers) {
                sb.append("\t\t<text ");
                sb.append(String.format("filename='%s' startRow='%d' startCol='%d' endRow='%d' endCol='%d'/>\n",
                                    textMarker.filename, textMarker.startRow, textMarker.startCol, textMarker.endRow, textMarker.endCol));
            }

            String comm = ai.getComment();
            if (comm != null)
            {
                if (comm.length() > 0)
                {
                    sb.append("\t\t<comment>");
                    if (!appstate.getUseCDATA())
                        sb.append(comm);
                    else
                    {
                        comm = comm.replaceAll("\\]\\]>", "] ]>");
                        sb.append(String.format("<![CDATA[%s]]>", comm));
                    }
                    sb.append("</comment>\n");
                }
            }
            sb.append("\t</annotation>\n");
        }
        sb.append("</annotations>\n");

        return sb.toString();
    }

}
