package sita;

import java.util.Enumeration;
import java.util.Hashtable;
import java.awt.Color;
import java.awt.BorderLayout;
import java.awt.Toolkit;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.KeyEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JFrame;
import javax.swing.UIManager;
import javax.swing.SwingUtilities;
import javax.swing.JSplitPane;
import javax.swing.AbstractAction;
import javax.swing.JPanel;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.KeyStroke;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import javax.swing.JOptionPane;
import javax.swing.WindowConstants;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.TabInfo;
import org.sanskritlibrary.Text;

/*
 * Main windows. Implements the following interfaces
 * 1. ActionListener (actionPerformed)
 *      passed to the FeatureMenuBar constructor so that all menu actions come here.
 * 2. ItemListener (itemStateChanged)
 *      also passed to the FeatureMenuBar constructor. Will be called when a menu drops down
 * 3. PropertyChangeListener (propertyChange)
 *      Called whenever position of JSplitPane divider is modified by user or when user changes orientation using Ctrl D
 * 4. Annotated (updateTitle, setColor, getColor, refreshManuscript, getDefaultDisplayWidth, setupInputActions, getAccelerator, allowModification)
 *      Allows other ApplicationState and other components to effect components defined in the main window.
 */
public class Main extends JFrame implements ActionListener, ItemListener, PropertyChangeListener, Annotated
{
    public static int actionDebug = 0;

    // *** Components used to build main window
    private FeaturesMenuBar menubar;
    private boolean filesTested = false;

    private ImagePanel img;             // display manuscript page
    private ImageScrollPane isp;        // allows the manuscript page to be scrolled
    private SearchArea sa;              // manages digital text and search of digital text
    private CommentScrollPage comment;    // comment part of an annotation
    private MarkedPanel commentPanel;      // to put red line to right of comment
    private JPanel rightPanel;          // contains sa and comment
    private JSplitPane splitPane;       // splitter between image and text
    private GoToDialog gtd;

    // *** Dialogs used by program. All of these are allocated only if they are needed
    private ChooseTextDialog ctd;       // Choose text dialog. Initially displayed if there is a manuscript root directory
    private ConfigurationDialog cd;     // Configuration dialog. Initially displayed if there is no manuscript root directory
    private AppearanceDialog appeard;   // Lets user change color of brackets and boxes used to mark up image
    private InformationDialog infod;    // Displays the full path of the current image file
    private ElizaDialog elizad;         // A joke. Will probably not be in version that is shipped to India
    private AnnotationChooserDialog annotationChooserd;
    private VerifyModificationDialog verifyd;   // A dialog that is displayed if the user attempts to modify a previously entered annotation.
    private VerseIndexDialog versed;
    private SummaryDialog summaryd;
    private CommentsDialog commentsd;
    private DifferencesDialog differencesd;
    private SearchTextDialog searchd;

    private ApplicationState appstate;  // A Singleton facade that maintains the global state of the application. Its instance is passed to all components.
    private Dimension setupDimensions;  // Dimensions of main window. Defined in constructor and used in displayManuscript.
    private Hashtable<String, AcceleratorMap> accelerators; // Map's menu hot keys to actions

    public static void main(final String[] args)
    {
        try
        {
            if (System.getProperty("os.name").equals("Mac OS X"))
                UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
        catch (Exception ex)
        {

        }
       //Schedule a job for the event dispatch thread: creating and showing this application's GUI.
	SwingUtilities.invokeLater(new Runnable()
    {
            public void run()
            {
                //Turn off metal's use of bold fonts
                UIManager.put("swing.boldMetal", Boolean.FALSE);
                try
                {
                    new Main(args).setVisible(true);
                }
                catch (Exception ex)        // message will be displayed in a popup before getting here.
                {
                    Utils.displayError(ex);
                }
            }
        });
    }
    
    public Main(String[] args) throws Exception
    {
        // appstate is switchboard that has knowledge all configuration files and the current state of the application
        appstate = new ApplicationState(this);

       /*
       TextAnalyzer ta = new TextAnalyzer(appstate);
       ta.fixupShift();
       System.exit(0);
       */
        
        setupAccelerators();    // for menu commands

        initComponents();

        // Set up polling for the number of annotations every second so that the annotation count on title bar is correct.
        /* Removed because of race condition concerns. Now only detect marks on image.
        int delay = 1000; // delay for 5 sec.
        int period = 1000; // repeat every sec.
        Timer timer = new Timer();
        UpdateTitleTask utt = new UpdateTitleTask(this);
        timer.scheduleAtFixedRate(utt, delay, period);
        */

        // If this is first time program is being run, display configuration dialog so manuscript root directory can be set.
        if (appstate.getManuscriptRootDirectory().equals("?"))
            configure();

        requestFocus();
    }


    /*
        Put information on each accelerator key in a hash table. This information is used by both
        menu bar and the search, digital text and comments textareas. The textareas are involved
        because the Macintosh gives the accelerator keys to the textarea that has the focus. Therefore,
        text areas and fields have to know what to do with an accelerator key.
    */
    private void setupAccelerators()
    {
        accelerators = new Hashtable<String, AcceleratorMap>();
        
        // Texts menu
        accelerators.put("T", new AcceleratorMap(KeyEvent.VK_T, new ChooseTextAction()));         // next page
        accelerators.put("N", new AcceleratorMap(KeyEvent.VK_N, new NextPageAction()));         // next page
        accelerators.put("P", new AcceleratorMap(KeyEvent.VK_P, new PrevPageAction()));         // previous page
        accelerators.put("S", new AcceleratorMap(KeyEvent.VK_S, new SaveDefaultsAction()));     // save defaults

        // View menu
        accelerators.put("D", new AcceleratorMap(KeyEvent.VK_D, new RotateDisplayAction()));    // rotate display
        accelerators.put("+", new AcceleratorMap(KeyEvent.VK_EQUALS, new ZoomInAction()));      // zoom in
        accelerators.put("-", new AcceleratorMap(KeyEvent.VK_MINUS, new ZoomOutAction()));      // zoom out
        accelerators.put("R", new AcceleratorMap(KeyEvent.VK_R, new RefreshScreenAction()));    // refresh screen
        accelerators.put("I", new AcceleratorMap(KeyEvent.VK_I, new FileInformationAction()));  // file information
        accelerators.put("B", new AcceleratorMap(KeyEvent.VK_B, new ShowBookmarksAction()));    // show bookmarks
        accelerators.put("G", new AcceleratorMap(KeyEvent.VK_G, new GoToLineAction()));    // show bookmarks

        // Annotations menus
        accelerators.put("A", new AcceleratorMap(KeyEvent.VK_6, new NewAnnotationAction()));    // new annotation
        accelerators.put("1", new AcceleratorMap(KeyEvent.VK_1, new SearchAreaAction()));       // choose search area
        accelerators.put("2", new AcceleratorMap(KeyEvent.VK_2, new TextAreaAction()));         // choose manuscript
        accelerators.put("3", new AcceleratorMap(KeyEvent.VK_3, new CommentsAreaAction()));     // comments area
        accelerators.put(">", new AcceleratorMap(KeyEvent.VK_PERIOD, new NextAnnotationAction()));   // next annotation
        accelerators.put("'", new AcceleratorMap(KeyEvent.VK_QUOTE, new ChooseAnnotationAction()));   // choose annotation
        accelerators.put("X", new AcceleratorMap(KeyEvent.VK_BACK_SPACE, new RemoveAnnotationAction()));      // remove annotation
        accelerators.put("[", new AcceleratorMap(KeyEvent.VK_OPEN_BRACKET, new SwapBracketsAction()));      // Swap start and end brackets
        accelerators.put("/", new AcceleratorMap(KeyEvent.VK_SLASH, new CannedCommentAction()));      // Swap start and end brackets
    }


    // Handles menu commands that do not have hot keys.
    public void handleCommand(String actionCommand)
    {
        try
        {
            // Texts menu (choose Manuscript, next page, previous page and save defaults all have hot keys and are handled elsewhere
            if (actionCommand.equals("configure"))
                configure();
            else if (actionCommand.equals("updateVirtual")) {
                appstate.updateVirtualFiles();
            }
            else if (actionCommand.equals("closeapp"))
            {
                saveState();
                System.exit(0);
            }
            // View menu  (rotate, zoom in, zoom out and refresh screen have hot keys and are handled elsewhere)
            else if (actionCommand.startsWith("encoding"))
                changeEncoding(actionCommand);
            else if (actionCommand.equals("increaseFontSize"))
                sa.increaseFontSize();
            else if (actionCommand.equals("decreaseFontSize"))
                sa.decreaseFontSize();
            else if (actionCommand.equals("appearance"))
                adjustAppearance();
            else if (actionCommand.equals("eliza"))
                displayEliza();

            // Annotations menu  (all commands have hot keys and are handled elsewhere
            // Analysis menu
            else if (actionCommand.equals("lookup"))
                _lookupVerse();
            else if (actionCommand.equals("annotationsreport"))
                _generateAnnotationsSummary();
            else if (actionCommand.equals("commentsreport"))
                _generateCommentsSummary();
            else if (actionCommand.equals("differencesreport"))
                _generateDifferencesSummary();
            else if (actionCommand.equals("searchtext"))
                _doSearchText();

            // Help menu
            else if (actionCommand.equals("help"))
                information("help");
            else if (actionCommand.equals("about"))
                information("about");
            else if (actionCommand.equals("reviewOnly")) {
                appstate.setReviewOnly(menubar.reviewOnly());
            }
        }
        catch (Exception ex)
        {
            Utils.displayError(ex);
            displayDefaultCursor();
        }

    }


    /***********************************************************************
     * The following are actions for menu commands that do not have hot keys.
     * They are called from the method handleCommand (see above)
     */

    // Displays the ChooseTextDialog
    private void chooseText() throws Exception
    {
        saveAnnotations();       // even if it hasn't changed.

        ctd.display();
        toFront();
        String filename = ctd.getChosenText();
        if (filename != null)
            displayManuscript(filename);
    }

    // Analyze Menu

    // Access by verse number instead of page number
    private void _lookupVerse()
    {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        try
                        {
                            lookupVerse();
                        }
                        catch (Exception ex)
                        {
                            Utils.displayError(ex);
                        }
                    }
               }
               );
    }


    // Access by verse number instead of page number
    private void lookupVerse() throws Exception
    {
        Utils.displayWarning("Under construction");
/*
        boolean currentManuscriptOnly = menubar.currentManuscriptOnly();
        saveAnnotations();
        appstate.createVerseIndex(currentManuscriptOnly);            // for current manuscript. do this every time in case new verses annotated since last time dialog was displayed.
        if (versed == null)
            versed = new VerseIndexDialog(appstate);
        versed.display(currentManuscriptOnly);
 */
    }


    // Analyze annotations based on 7 different annotation types, pull error and out of order annotations
    private void _generateAnnotationsSummary()
    {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        try
                        {
                            generateAnnotationsSummary();
                        }
                        catch (Exception ex)
                        {
                            Utils.displayError(ex);
                        }
                    }
               }
               );
    }


    // Analyze annotations based on 7 different annotation types, pull error and out of order annotations
    private void generateAnnotationsSummary() throws Exception
    {
        boolean currentManuscriptOnly = menubar.currentManuscriptOnly();
        currentManuscriptOnly = true;       // remove
        saveAnnotations();
        appstate.generateAnnotationsSummary(currentManuscriptOnly);            // for current manuscript
        summaryd = new SummaryDialog(appstate);
        summaryd.display(currentManuscriptOnly);
    }

    // Look up pages with a particular comment in them
    private void _generateCommentsSummary()
    {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        try
                        {
                            generateCommentsSummary();
                        }
                        catch (Exception ex)
                        {
                            Utils.displayError(ex);
                        }
                    }
               }
               );
    }


    // Lookup pages with a particular comment in them
    private void generateCommentsSummary() throws Exception
    {
        boolean currentManuscriptOnly = menubar.currentManuscriptOnly();
        currentManuscriptOnly = true;       // remove
        saveAnnotations();
        appstate.generateCommentsSummary(currentManuscriptOnly);            // for current manuscript
        commentsd = new CommentsDialog(appstate);
        commentsd.display(currentManuscriptOnly);
    }


    // Compare current annotations with original ones.
    private void _generateDifferencesSummary()
    {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        try
                        {
                            generateDifferencesSummary();
                        }
                        catch (Exception ex)
                        {
                            Utils.displayError(ex);
                        }
                    }
               }
               );
    }


    // Compare current annotations with original ones.
    private void generateDifferencesSummary() throws Exception
    {
        boolean currentManuscriptOnly = menubar.currentManuscriptOnly();
        currentManuscriptOnly = true;       // remove
        saveAnnotations();
        differencesd = new DifferencesDialog(appstate);
        differencesd.display(currentManuscriptOnly);
    }


    private void _doSearchText()
    {
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        try
                        {
                            doSearchText();
                        }
                        catch (Exception ex)
                        {
                            Utils.displayError(ex);
                        }
                    }
               }
               );
    }


    // Compare current annotations with original ones.
    private void doSearchText() throws Exception
    {
        Utils.displayWarning("Needs to be convered to work with tabs");
/*
        boolean currentManuscriptOnly = menubar.currentManuscriptOnly();
        saveAnnotations();
        searchd = new SearchTextDialog(appstate);
        searchd.display(currentManuscriptOnly);
 */
    }

    // Displays the configuration dialog
    private void configure()
    {
        if (cd == null)
            cd = new ConfigurationDialog(appstate);

        cd.setVisible(true);
    }


    // Changes the encoding of digital texts. Choices are SLP1, Devanagari, Roman transliteration, Telugu
    private void changeEncoding(String actionCommand) throws Exception
    {
        saveState();                        // in case annotations on current page have changed.
        String newEncoding = actionCommand.substring(8);
        if (appstate.getEncoding().equals(newEncoding))
            Utils.displayWarning(String.format("Digital text is already encoded as %s", newEncoding));
        else if (appstate.haveSelectedText())
        {
            String msg = String.format("You have requested that the digital text encoding be changed from %s to %s.\n\nAny existing text annotations will be adjusted for the new encoding.\n\nPlease backup your %s directory before proceeding.\n\nReady?",
                    appstate.getEncoding(), newEncoding, appstate.getAnnotationsCurrentDirectory());
            int response  = JOptionPane.showConfirmDialog(this, 
                    msg,
                    "Change Encoding",
                    JOptionPane.YES_NO_OPTION);

            if (response == JOptionPane.YES_OPTION)
            {
                StringBuilder result = new StringBuilder();
                boolean reformatSuccessful = appstate.reformat(appstate.getEncoding(), newEncoding, result);
                if (reformatSuccessful)
                    appstate.setEncoding(newEncoding);
                appstate.forceReloadOfAnnotations();
                displayManuscript(ctd.getChosenText());                   // reload the digital text with the current encoding.
                if (reformatSuccessful)
                {
                    String message = null;
                    if (result.length() == 0)
                        message = "Annotation files reformatted";
                    else
                    {
                        String filename = String.format("%s.adjustments.txt", appstate.getLastManuscript().replaceAll("[\\\\/]", "_"));
                        try
                        {
                            BufferedWriter bw = IOAssistance.openWriter(filename);
                            bw.write(result.toString());
                            bw.close();
                            File f = new File(filename);
                            message = String.format("Annotation files reformatted with adjustments<br/> For details, see %s", f.getAbsolutePath());
                        }
                        catch (Exception ex)
                        {
                            String s = result.toString().replaceAll("\n", "<br/>");
                            Utils.displayWarning("Annotation files reformatted with adjustments<br/>" + s);
                        }
                    }

                    if (message != null)
                        Utils.displayWarning(message);
                }
            }
        }
        else
        {
            appstate.setEncoding(newEncoding);
            sa.loadTabs();                   // reload the digital text with the new encoding.
        }
    }

    // Display dialog that lets user change color of brackets or box on manuscript image
    public void adjustAppearance()
    {
        if (appeard == null)
            appeard = new AppearanceDialog(appstate, this);

        appeard.initialize();
        appeard.setVisible(true);
    }


    // A joke. Probably won't be in version that gets sent to India.
    public void displayEliza()
    {
        if (elizad == null)
            elizad = new ElizaDialog();
        elizad.display();
    }



//http://java.itags.org/java-socket-programming/240464/
    private void initComponents() throws Exception
    {
        JFrame.setDefaultLookAndFeelDecorated(false);
        this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        if (System.getProperty("os.name").equals("Mac OS X"))
        {
            OSXAdapter.setQuitHandler(this, getClass().getDeclaredMethod("quit", (Class[])null));
        } else // if (appstate.getCanCloseWindow())
        {
            addWindowListener(new WindowAdapter()      // what to do when user shuts down app
            {

                @Override
                public void windowClosing(WindowEvent e)
                {
                    saveState();
                    System.exit(0);
                }
            });
        }

       menubar = new FeaturesMenuBar(appstate, this, this);         // set up the menu bar and its accelerators

       Toolkit tk = Toolkit.getDefaultToolkit();
       Dimension d = tk.getScreenSize();
       int defaultWidth = (int)(d.getWidth()-50);
       setupDimensions = new Dimension(defaultWidth, (int)(d.getHeight()-50));    // window fills screens except of 50 pixels of width and 50 pixels of height
       this.setJMenuBar(menubar);
       this.setSize((int)setupDimensions.getWidth(), (int)setupDimensions.getHeight());
       this.setTitle("Manuscript Correlation");
       this.setLayout(new BorderLayout());

       //displayManuscript(null);

       ctd = new ChooseTextDialog(appstate);

       // If manuscript root directory has been specified, display ChooseTextDialog so user can resume annotating.
       if (!appstate.getManuscriptRootDirectory().equals("?"))
       {
           ctd.display();
           String filename = ctd.getChosenText();
            if (filename != null)
                displayManuscript(filename);        // display the chosen image and its digital text.
       }
    }


    // Called when user closes application, either by clicking x or choosing Quit from text menu.
    public void saveState()
    {
        try
        {
            saveAnnotations();      // make sure last annotation is saved
        } catch (Exception ex)
        {
            Logger.writeLog(ex);
        }
        try
        {
            appstate.saveState();               // so that ChooseTextDialog will display current page next time program is run.
        } catch (Exception ex)            // Hopefully, user has been warned about this earlier.
        {
            Logger.writeLog(ex);
        }
    }

    public void quit()
    {
        saveState();
        System.exit(0);
    }

    // This is a workhorse. It is called whenever the user requests a new page.
    public void displayManuscript(String imageFile) throws Exception
    {
        Dimension minimumSize = new Dimension(100, 50);

        boolean newManuscript = appstate.updateLocalVersionsOfDefaults();   // in case manuscript has changed, get new defaults.
        appstate.setAllowModification(true);        // since the empty annotation is initially displayed, let the user annotate it.
        if (imageFile != null)
        {
            Utils.displayWaitCursor(this);

            if (img == null)            //  Set up scrolled image controls
            {
                int displayWidth = 0;
                int displayHeight = 0;
                int dividerLocation = getDefaultDivider();
                if (appstate.getDefaultOrientation() == JSplitPane.HORIZONTAL_SPLIT)  // (one to the left of the other)
                {
                    displayWidth = dividerLocation;
                    displayHeight = this.getHeight()-50;
                }
                else        // JSplitPane.VERTICAL_SPLIT (one on top of one another)
                {
                    displayWidth = (this.getWidth()-50);
                    displayHeight = dividerLocation;
                }
                img = new ImagePanel(appstate, displayWidth, displayHeight);
                isp = new ImageScrollPane(img);
                isp.setMinimumSize(minimumSize);
            }

            try
            {
                appstate.getAnnotations();      // load any previously made annotations for this page
                img.loadImage(imageFile);           // load the image into the scrolled image control
                img.clearMarkerData();
            }
            catch (Exception ex)
            {
                displayDefaultCursor();
                Utils.displayError(ex);
                appstate.setDisplaySuccessful(false);
                img = null;
                return;
            }

            if (!appstate.isDisplaySuccessful())     // if the app was just started and this is the first image.
            {
                menubar.enableCommands();

                // create area that will contain searchable digital text
                if (sa == null)
                    sa = new SearchArea(appstate, minimumSize, this);

                // create comments area below the digital text area.
                if (comment == null)
                {
                    Dimension size = new Dimension(this.getWidth()/2, 50);
                    comment = new CommentScrollPage(appstate, 4, sa.getTextCols(), size);
                    commentPanel = new MarkedPanel(comment);
                    comment.setMarkedPanel(commentPanel);
                }

                // use a BorderLayout to layout the searchable text and the comment area
                if (rightPanel == null)
                {
                    rightPanel = new JPanel();
                    rightPanel.setLayout(new BorderLayout());
                    rightPanel.add(sa, BorderLayout.CENTER);
                    rightPanel.add(commentPanel, BorderLayout.SOUTH);
                }

                boolean doPack = true;
                if (splitPane == null)
                {
                    splitPane = new JSplitPane(appstate.getDefaultOrientation(), isp, rightPanel);
                    splitPane.setDividerLocation(getDefaultDivider());
                    splitPane.setOneTouchExpandable(true);
                    splitPane.addPropertyChangeListener(JSplitPane.DIVIDER_LOCATION_PROPERTY, this);
                    this.add(splitPane, BorderLayout.CENTER);       // everything is in the splitPane
                }
                else
                {
                    if (appstate.getOrientation() == JSplitPane.HORIZONTAL_SPLIT)
                        splitPane.setLeftComponent(isp);
                    else
                        splitPane.setTopComponent(isp);
                    splitPane.validate();
                    doPack = false;
                }

                if (doPack)
                    this.pack();
                this.setSize(setupDimensions.width, setupDimensions.height);
                this.setVisible(true);

                appstate.setDisplaySuccessful(true);        // so that this block is not executed again.
            }

            img.clearTemporaryMarks();

            int ddl = getDefaultDivider();

            // The following test of orientation **is** necessary if user changes from manuscript with one default orientation to one with another default orientation
            if (appstate.getDefaultOrientation() != splitPane.getOrientation())
            {
                splitPane.setOrientation(appstate.getDefaultOrientation());
                splitPane.setDividerLocation(ddl);
            }

            // The following test of the divider location **is** necessary because the user might change the size of the main window!
            if (ddl != splitPane.getDividerLocation())
                splitPane.setDividerLocation(ddl);

            if (newManuscript) {
                sa.loadTabs();       // in case user is viewing a new manuscript with a different digital text. Usually not the case.
            }

            menubar.enableCommands();       // so that hot keys work correctly

            // The current annotation is the empty one. Displaying it will effectively remove any old annotations
            _displayAnnotation();

            appstate.denormalizeAnnotations(img.getActualWidth());

            this.setTitle(ctd.getProgress());

            displayDefaultCursor();   // get rid of wait cursor

            if (appstate.getMoveToAnnotationNumber() != -1)
                jumpToAnnotationNumber();
        }

        if (!filesTested)
        {
            filesTested = true;
            String msg = Utils.testWritableFile("log.txt");
            if (msg != null)
                Utils.displayWarning(String.format("An attempt to write to log.txt caused the following error message:<br/>%s",msg));

            msg = Utils.testWritableFile("state.txt");
            if (msg != null)
                Utils.displayWarning(String.format("An attempt to write to state.txt caused the following error message:<br/>%s", msg));
        }

        //unitTest();
    }


    public void unitTest() {
        String[] _lines = new String[]{
            "0 Comment",
            "1 Comment",
            "2 f1.0",
            "3 f1.1",
            "4 f1.2 xxxxx[xxx",
            "5 f1.3",
            "6 f1.4",
            "7 f1.5 ......",
            "8  Comment",
            "9  Comment",
            "10 f2.0",
            "11 f2.1",
            "12 f2.2",
            "13 f2.3",
            "14 f2.4",
            "15 f2.5 ..........",
            "16 Comment",
            "17 Comment",
            "18 f3.0",
            "19 f3.1",
            "20 f3.2 xxxxxxxxxxx]xxxx",
            "21 f3.3",
            "22 f3.4",
            "23 f3.5",};
        ArrayList<String> lines = new ArrayList<String>();
        for (String line : _lines) {
            lines.add(line);
        }

        ArrayList<TextMap> textmap = new ArrayList<TextMap>();
        textmap.add(new TextMap(Text.Type.Comment, null, 0, 1, 0, 0));        // tm0
        textmap.add(new TextMap(Text.Type.TextRef, "f1", 2, 7, 0, 0));        // tm1
        textmap.add(new TextMap(Text.Type.Comment, null, 8, 9, 0, 0));        // tm2
        textmap.add(new TextMap(Text.Type.TextRef, "f2", 10, 15, 0, 0));      // tm3
        textmap.add(new TextMap(Text.Type.Comment, null, 16, 17, 0, 0));      // tm4
        textmap.add(new TextMap(Text.Type.TextRef, "f3", 18, 23, 0, 0));      // tm5

        TextPoint annotationStart = new TextPoint(4, 10);
        //TextPoint annotationEnd = new TextPoint(20, 15);
        TextPoint annotationEnd = new TextPoint(6, 5);

        appstate.clearAnnotations();
        sa.dtps.get(0).generateTextMarkers(annotationStart, annotationEnd, textmap, lines);
        System.out.println(appstate);
    }
/*
tm0 type=TEXT filename=null, startrow=0, endrow=1
tm1 type=TEXTREF, filename=f1, startrow=2, endrow=7
tm2 type=TEXT filename=null, startrow=8, endrow=9
tm3 type=TEXTREF, filename=f2, startrow=10, endrow=15
tm4 type=TEXT, filename=null, startrow=16, endrow=17
tm5 type=TEXTREF, filename=f3, startrow=18, endrow=23
 */
    /*****************************************************
     * Private helper methods
     *****************************************************/

    // Sets the current annotation to ai and displays it.
    private void displayAnnotation(AnnotationInfo ai)
    {
        if (ai != null)
        {
            appstate.setCurrentAnnotationInfo(ai);

            _displayAnnotation();
        }
    }


    // display the current annotation
    private void _displayAnnotation()
    {
            refreshManuscript();
            displayComment();    // in an unfortunate dependency, this must be done before sa.display(). See appstate.isCurrentAnnotationEmpty() for the reason.
            sa.clearAnnotation();   // in case we are displaying the empty annotation after deleting the last existing annotation.
            sa.selectTab();         // make sure the right tab is visible if tabs exist.
            sa.display();
    }


    // Calculation is based on orientation
    private int getDefaultDivider()
    {
        if (appstate.getDefaultOrientation() == JSplitPane.VERTICAL_SPLIT)
            return (int)(appstate.getDefaultDivider() * this.getHeight());
        else
            return (int)(appstate.getDefaultDivider() * this.getWidth());
    }


    /* Write current annotations to disk. Is called in the following situations:
     *      1. New page chosen
     *      2. Ctrl N  display next page
     *      3. Ctrl P  display previous page
     *      4. Main window is closed
     */
    public void saveAnnotations()
    {
        try
        {
            if (img != null)
            {
                appstate.getCurrentAnnotation().setTabNumber(appstate.getCurrentTabNumber());
                appstate.getCurrentAnnotation().setEncoding(appstate.getEncoding());
                appstate.setComment(getComment());      // get what is currently displayed in comment field and save it in currentAnnotation
                sa.saveSelection();
                appstate.saveAnnotations(img.getResolvedImageWidth(true), img.getActualWidth());
            }
        }
        catch (Exception ex)
        {
            Utils.displayError(ex);
        }
    }

    // Display comment portion of annotation in comment area in lower right corner of window.
    private void displayComment()
    {
        comment.display();
    }

    /************************************************************************
     * End of private helper methods
     ************************************************************************/

    
    /************************************************
     * Required because this class implements ActionListener
     * Called whenever a menu command is selected by user (see FeatureMenuBar)
     */
    public void actionPerformed(ActionEvent e)
    {
        actionDebug++;

        AcceleratorMap map = accelerators.get(e.getActionCommand());
        if (map != null)         // if action has a hot key
            map.action.actionPerformed(null);
        else
            handleCommand(e.getActionCommand());
    }


    /***************************************************
     * Required because this class implements ItemListener
     * Called whenever a menu is pulled down so that commands in the menu can be enabled or disabled (See FeatureMenuBar)
     ***************************************************/
    public void itemStateChanged(ItemEvent e)
    {
        menubar.enableCommands();
    }



    /**********************************************************
     * Required because this class implements PropertyChangeListener
     * Called whenever position of JSplitPane divider is modified by user or when user changes orientation using Ctrl D
     */
    public void propertyChange(PropertyChangeEvent evt)
    {
        int newValue = (Integer)evt.getNewValue();

        if (splitPane.getOrientation() == JSplitPane.HORIZONTAL_SPLIT)   // left and right
        {
            appstate.setSplitDivider(newValue, this.getWidth());
            img.setDisplayWidth(newValue);          // so that clipping is correct
            img.setDisplayHeight(this.getHeight() - 50);
        }
        else        // top and bottom
        {
            appstate.setSplitDivider(newValue, this.getHeight());
            img.setDisplayWidth(this.getWidth() - 50);
            img.setDisplayHeight(newValue);         // so that clipping is correct
        }
    }



    /***************************************************************************
    * The following methods are required because this class implements Annotated.
    * The Annotated interface allows methods in ApplicationState to call into this
    * class in a reasonably elegant way.
    ****************************************************************************/
    
    // Keeps the annotation count somewhat up to date
    public void updateTitle()
    {
        if (appstate.isDisplaySuccessful())
        {
            //sa.saveSelection();               too dangerous
            //comment.saveSelection();
            this.setTitle(ctd.getProgress());
        }
    }


    // Called by AppearanceDialog
    public void setColor(Color c)
    {
        img.setPointColor(c);
        appstate.setPointColor(c);
    }

    // Called by AppearanceDialog
    public Color getColor()
    {
        return img.getPointColor();
    }


    // Called by the ApplicationState methods saveAnnotations and isCurrentAnnotationEmpty
    public String getComment()
    {
        return comment.getText();
    }


    // Called by RubberBand and RubberBandRectangle when displays needs to be cleared.
    public void refreshManuscript()
    {
        img.display();
    }

    public void refreshManuscript(boolean adjustScrolling)
    {
        img.display(adjustScrolling);
    }

    // Called by ApplicationState.getDefaultWidth if there was no default element in config.xml for the current manuscript.
    public int getDefaultDisplayWidth()
    {
        return (this.getWidth() - 50)/2;
    }

    // Called by VerseIndexDialog to change to the one selected by the user.
    public void setCurrentPage(String manuscriptDirectory, String pageName)
    {
        ctd.setCurrentPage(manuscriptDirectory, pageName);
    }

    public void reloadAnnotations(String currentDirname, String currentFilename) throws Exception
    {
        appstate.annotations.loadAnnotations(currentDirname, currentFilename);
        appstate.denormalizeAnnotations(img.getActualWidth());
    }

    public void bringToFront()
    {
        toFront();
    }

    // Make each menu accelerator an input action. Called by CommentTextArea, SearchArea and DigitalTextPage.
    // This is not necessary on Windows but is necessary on the Mac because accelerator keys appears to be
    // sent to the component which has the focus which then has to pass it on.
    public void setupInputActions(InputMap im, ActionMap am)
    {
        Enumeration<String> e = accelerators.keys();
        while (e.hasMoreElements())
        {
            String key = e.nextElement();
            AcceleratorMap map = accelerators.get(key);
            im.put(map.k, key);
            am.put(key, map.action);
        }
    }


    // Called by FeaturesMenuBar to get the same accelerators that we sent to the textboxes.
    // This design allows me to specify the accelerator keys and their actions in only one place.
    public KeyStroke getAccelerator(String key)
    {
        AcceleratorMap am = accelerators.get(key);
        if (am == null)
            Utils.displayError(String.format("%s is not a known accelerator", key));
        return am.k;
    }


    // The user has tried to make a modification to an existing annotation. This method
    // pops up a dialog box that gives the user the change to create a new annotation instead,
    // modify the existing annotation or do nothing. Called by CommentTextArea (user has
    // tried to modify a comment), RubberBand (user has tried to modify a marker or box),
    // DigitalTextPage (user has tried to modify the digital text highlight) or from
    // this class if user has done a Ctrl T or Ctrl C.
    public boolean allowModification()
    {
        boolean rc = false;

        if (verifyd == null)
            verifyd = new VerifyModificationDialog(appstate);

        verifyd.display();

        this.toFront();     // for Ubuntu

        if (verifyd.getCreateNewAnnotation())
        {
            newAnnotation(false);
            rc = true;
        }
        else
            rc = appstate.getAllowModification();

        return rc;
    }

    // Called by ImagePanel because it is an Image
    public void displayDefaultCursor()
    {
        Utils.displayDefaultCursor(this);
    }

    public void debug(String s)
    {
        comment.debug(s);
    }

    public void debugAppend(String s)
    {
        comment.debugAppend(s);
    }

    /***************************************************************************
     * End of methods required by Annotated interface
     ***************************************************************************/

    /****************************************************************************************
     * The methods and classes implement menu actions. The classes extend AbstractAction which
     * has an abstract method actionPerformed. Each class has a corresponding local method
     * that is called by the actionPerformed method of the class. I did things this way to
     * make it easy to programmatically do the action without requiring the user to request it.
     *****************************************************************************************/


    private class ChooseTextAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            try
            {
                chooseText();
            }
            catch (Exception ex)
            {
                Utils.displayError(ex);
                displayDefaultCursor();
            }
        }
    }

    // Ctrl + N  Next page
    private void nextText()
    {
        try
        {
            saveAnnotations();       // even if it has not changed

            String filename = ctd.getNextText();
            if (filename != null) {
                displayManuscript(filename);
            } else
                Utils.displayWarning("At last page");
            }
            catch (Exception ex)
            {
                Utils.displayError(ex);
            }
    }
    private class NextPageAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            nextText();
        }
    }


    // Ctrl P  Previous page
    private void prevText()
    {
        try
        {
            saveAnnotations();       // even if it has not changed.

            String filename = ctd.getPrevText();
            if (filename != null) {
                displayManuscript(filename);
            }
            else
                Utils.displayWarning("At first page");
        }
        catch (Exception ex)
        {
            Utils.displayError(ex);
        }
    }
    private class PrevPageAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            prevText();
        }
    }


    // Ctrl S  Save defaults
    public void saveDefaults()
    {
        try
        {
            int maxdiv = this.getWidth();
            if (splitPane.getOrientation() == JSplitPane.VERTICAL_SPLIT)
                maxdiv = this.getHeight();
            appstate.saveDefaults(splitPane.getDividerLocation(), maxdiv,
                    img.getResolvedImageWidth(), appstate.getPointColor(),
                    splitPane.getOrientation());
        }
        catch (Exception ex)
        {
            Utils.displayError(ex);
        }
    }
    private class SaveDefaultsAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            saveDefaults();
        }
    }


    // Ctrl D  Rotate display
    private void rotateDisplay()
    {
        int dividerOffset = 0;

        if (splitPane.getOrientation() == JSplitPane.HORIZONTAL_SPLIT)
        {
            splitPane.setOrientation(JSplitPane.VERTICAL_SPLIT);
            dividerOffset = getHeight() / 2;
            img.setDisplayWidth(getWidth() - 50);
            img.setDisplayHeight(getHeight() / 2);
        } else
        {
            splitPane.setOrientation(JSplitPane.HORIZONTAL_SPLIT);
            dividerOffset = getWidth() / 2;
            img.setDisplayWidth(getWidth() / 2);
            img.setDisplayHeight(getHeight() - 50);
        }
        splitPane.setDividerLocation(dividerOffset);
        appstate.setOrientation(splitPane.getOrientation());
        refreshManuscript();

    }
    private class RotateDisplayAction extends AbstractAction
    {

        public void actionPerformed(ActionEvent ev)
        {
            rotateDisplay();
        }
    }


    // Ctrl =  Zoom in
    private void zoomIn()
    {
        appstate.setUseDefaultWidth(false);
        int width = img.getResolvedImageWidth();
        img.setResolvedImageWidth(width + 100);
        refreshManuscript();
    }


    private class ZoomInAction extends AbstractAction
    {

        public void actionPerformed(ActionEvent ev)
        {
            zoomIn();
        }
    }



    // Crtl -  Zoom out
    private void zoomOut()
    {
        appstate.setUseDefaultWidth(false);
        int width = img.getResolvedImageWidth();
        if (width > 200)
        {
            img.setResolvedImageWidth(width - 100);
            refreshManuscript();
        }
    }


    private class ZoomOutAction extends AbstractAction
    {

        public void actionPerformed(ActionEvent ev)
        {
            zoomOut();
        }
    }


    // Ctrl R  (The refreshManuscript method is a member of the Annotated interface and appears with other Annotated methods

    private class RefreshScreenAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
                refreshManuscript();
        }
    }


    // Ctrl I  Show file information

    public void information(String type)
    {
        if (infod == null)
            infod = new InformationDialog(appstate);
        infod.showInfo(type);
        infod.setVisible(true);
    }


    private class FileInformationAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
                information("file");
        }
    }


    private class ShowBookmarksAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            ArrayList<BookmarkInfo> bookmarks = new ArrayList<BookmarkInfo>();
            ArrayList<TextMap> textmap = appstate.getTextMap();
            ArrayList<String> lines = appstate.getTextLines();

            TextPoint tp = sa.getEndOfPreviousAnnotation();
            if (tp != null) {
                bookmarks.add(new BookmarkInfo("End of previous annotation", tp.row));
            }

            tp = sa.getStartOfCurrentAnnotation();
            if (tp != null) {
                bookmarks.add(new BookmarkInfo("Start of current annotation", tp.row));
            }

            appstate.getBookmarks(bookmarks, textmap, lines);
            BookmarkDialog bd = new BookmarkDialog(Main.this, bookmarks);
            bd.setVisible(true);
            int row = bd.getBookmarkOffset();
            sa.setScroller(row);
        }
    }

    private class GoToLineAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            if (gtd == null) {
                gtd = new GoToDialog(Main.this);
            }
            gtd.display();
            if (gtd.getJumpToLine()) {
                int row = gtd.getLine();
                sa.setScroller(row);
            }
        }
    }

    // Ctrl A  New annotation
    public void newAnnotation(boolean onlyIfTextIsSelected) {
        boolean createNewAnnotation = true;
        if (onlyIfTextIsSelected) {
            createNewAnnotation = sa.selectionExists();
        }

        if (createNewAnnotation) {
            try {
                // remember which tab annotation ended up in
                appstate.getCurrentAnnotation().setTabNumber(appstate.getCurrentTabNumber());
                appstate.getCurrentAnnotation().setEncoding(appstate.getEncoding());

                appstate.setAllowModification(true);
                appstate.setComment(getComment());      // get what is currently displayed in comment field and save it in currentAnnotation
                sa.saveSelection();
                appstate.resetCurrentAnnotation();      // make the current annotation an empty one

                // The current annotation is now the empty one. Displaying it will remove any traces of the previous annotation.
                _displayAnnotation();

                setTitle(ctd.getProgress());
            } catch (Exception ex) {
                Utils.displayError(ex);
            }
        }
    }

    
    private class NewAnnotationAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            newAnnotation(false);
        }
    }


    // Ctrl 1  Search area
    public void moveToSearchArea()
    {
        sa.setSearchAreaFocus();
    }

    private class SearchAreaAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            moveToSearchArea();
        }
    }


    // Ctrl 2  Text area
    public void moveToTextArea()
    {
        if (appstate.getAllowModification() || allowModification())
            sa.setTextFocus();
    }
    private class TextAreaAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            moveToTextArea();
        }
    }


    // Ctrl 3  Comments area
    public void moveToCommentsArea()
    {
        if (appstate.getAllowModification() || allowModification())
            comment.setFocus();
    }
    private class CommentsAreaAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            moveToCommentsArea();
        }
    }


    public void jumpToAnnotationNumber()
    {
        int n = appstate.getMoveToAnnotationNumber();         // set by CommentsDialog, et al
        appstate.setMoveToAnnotationNumber(-1);
        
        appstate.setComment(getComment());      // get what is currently displayed in comment field and save it in currentAnnotation
        sa.saveSelection();

        int numAnnotations = appstate.getNumAnnotations();
        if (numAnnotations > 0 && n < numAnnotations)
        {
            for (int i=0; i<numAnnotations; i++)
            {
                if (!appstate.getAnnotation(i).isEmpty())
                {
                    if (i == n)
                    {
                        appstate.setAllowModification(false);
                        AnnotationInfo ai = appstate.getAnnotation(n);
                        displayAnnotation(ai);                        
                        n++;
                        if (n == numAnnotations)
                            n = 0;
                        appstate.setNextAnnotationNumber(n);
                        break;
                    }
                }
            }
        }
        else
            appstate.setAllowModification(true);

        updateTitle();
        
    }


    // Ctrl >  Next annotation
    public void nextAnnotation()
    {
        appstate.setComment(getComment());      // get what is currently displayed in comment field and save it in currentAnnotation
        sa.saveSelection();

        int numAnnotations = appstate.getNumAnnotations();
        if (numAnnotations > 0)
        {
            appstate.setAllowModification(false);
            int n = appstate.getNextAnnotationNumber();
            if (n >= numAnnotations)        // was an annotation removed?
                n = numAnnotations - 1;

            AnnotationInfo ai = appstate.getAnnotation(n);
            displayAnnotation(ai);

            n++;

            if (n == numAnnotations)
            {
                java.awt.Toolkit.getDefaultToolkit().beep();
                n = 0;
            }

            appstate.setNextAnnotationNumber(n);
        }
        else
            appstate.setAllowModification(true);

        updateTitle();
    }
    
    private class NextAnnotationAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            nextAnnotation();
        }
    }

    private class ChooseAnnotationAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            appstate.setComment(getComment());      // get what is currently displayed in comment field and save it in currentAnnotation
            sa.saveSelection();

            if (annotationChooserd == null)
                annotationChooserd = new AnnotationChooserDialog(appstate);
            annotationChooserd.display();
            int selectedAnnotation = annotationChooserd.getAnnotation();
            if (selectedAnnotation != -1)
                appstate.setAllowModification(false);
                displayAnnotation(appstate.getAnnotation(selectedAnnotation));
        }
    }


    // Ctrl Backspace  Remove annotation
    public void removeAnnotation()
    {
        appstate.removeCurrentAnnotation();

        // The current annotation should now be the empty annotation. Displaying it will remove the just deleted annotation.
        _displayAnnotation();

        appstate.setAllowModification(true);

        updateTitle();
    }
    
    private class RemoveAnnotationAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            removeAnnotation();
        }
    }

    public void swapBrackets()
    {
        img.swapBrackets();
    }

    private class SwapBracketsAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            swapBrackets();
        }
    }

    public void insertCannedComment()
    {
        comment.insertCannedComment();
    }

    private class CannedCommentAction extends AbstractAction
    {
        public void actionPerformed(ActionEvent ev)
        {
            insertCannedComment();
        }
    }

}
