package sita;

import java.awt.Point;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.CatalogRecord;
import org.sanskritlibrary.FSM;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.ResourceInputStream;
import org.sanskritlibrary.ResourceInputStreamImpPath;
import org.sanskritlibrary.webservice.TransformMap;
import org.sanskritlibrary.webservice.WebServices;

public class TextAnalyzer
{
    public static final int typeCommentOnly = 0;
    public static final int typeMarkersOnly = 1;
    public static final int typeHighlightOnly = 2;
    public static final int typeCommentMarkers = 3;
    public static final int typeCommentHighlight = 4;
    public static final int typeMarkersHighlight = 5;
    public static final int typeCommentMarkersHighlight = 6;
    public static final int typeError = 7;
    public static final int typeOutOfOrder = 8;
    public static final int typeLast = typeOutOfOrder;

    // The following needs to be kept in sync with the above constants.
    public static String[] types = new String[]
    {
        "Comment only",
        "Markers only",
        "Highlight only",
        "Comment and markers",
        "Comment and highlight",
        "Markers and highlight",
        "Comment, markers and highlight",
        "Error",
        "Out of order",
    };

    Hashtable<String, FSM[]> fsmh = new Hashtable<String, FSM[]>();
    ResourceInputStream ris = new ResourceInputStreamImpPath("transcoders");

    ApplicationState appstate;
    private int eolLength = 0;
    private String oldText;
    private String[] oldLines;
    private String newText;
    private String[] newLines;
    private Hashtable<String, String[][]> textLines = new Hashtable<String, String[][]>();
    private String oldEncoding;
    private String newEncoding;
    private StringBuilder sb = new StringBuilder();
    private TransformMap tmOldNew;
    private TransformMap tmNewSLP1;
    private TransformMap tmSLP1Old;

    // for round trip
    private TransformMap _tmNewOld;
    private TransformMap _tmOldSLP1;
    private TransformMap _tmSLP1New;
    private TransformMap tm;

    ConversionContext forwardContext;
    ConversionContext returnContext;

    private Hashtable<String, ArrayList<VerseAnnotation>> verseIndex;

    public TextAnalyzer(ApplicationState appstate)
    {
        this.appstate = appstate;
    }


    public ArrayList<Hashtable<String, String>> generateAnnotationsSummary(String dirname, boolean useCurrentManuscript) throws Exception
    {
        Annotations annotations = new Annotations(appstate);
        ArrayList<Hashtable<String, String>> summary = new ArrayList<Hashtable<String, String>>();
        int commentOnly = 0;                // type = 1
        int markersOnly = 0;                // type = 2
        int highlightOnly = 0;              // type = 3
        int commentMarkers = 0;             // type = 4
        int commentHighlight = 0;           // type = 5
        int markersHighlight = 0;           // type = 6
        int commentMarkersHighlight = 0;    // type = 7
        String relativeDirname = null;

        for (int i = 0; i <= typeLast; i++)
        {
            summary.add(new Hashtable<String, String>());
        }

        Regex re = new Regex(dirname, String.format("%s/(.+)$", appstate.getAnnotationsRootDirectory()));
        if (re.find())
            relativeDirname = re.getGroup(1);
        String outfile = dirname + "/annotationSummary.txt";
        BufferedWriter bw = IOAssistance.openWriter(outfile);
        //String currentDirname = appstate.annotations.getDirname();
        //String currentFilename = appstate.annotations.getFilename();
        File dir = new File(appstate.getAnnotationsCurrentDirectory());
        File[] ff = dir.listFiles();
        boolean errorReported = false;
        String filekey = null;
        ArrayList<AnnotationInfo> sortList = new ArrayList<AnnotationInfo>();

        for (File f : ff)
        {
            String filename = f.getAbsolutePath();
            String lastpart = f.getName();
            int n = 0;
            if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
            {
                sortList.clear();
                Regex rex = new Regex(filename, "[\\\\/]([^\\\\/]+)\\.xml$");
                if (rex.find())
                    filekey = rex.getGroup(1);                  // this is used to associate summary information with this xml file
                else
                    filekey = filename;     // should never reach here.

                try
                {
                    annotations.loadAnnotations(relativeDirname, f.getAbsolutePath());         // these will be normalized
                }
                catch (Exception ex)
                {
                    Utils.displayError(ex);
                }

                ArrayList<AnnotationInfo> ais = annotations.getAnnotations();
                for (AnnotationInfo ai : ais)
                {
                    String msg = null;
                    n++;
                    ai.setSequenceNumber(n);
                    errorReported = false;

                    // Test this annotation for errors

                    // Verify that if there is a right bracket, then there is a left bracket.
                    if (!ai.getLeftBracketExists() && ai.getRightBracketExists())
                        msg = String.format("annotation %d. Right bracket with no left bracket", n);

                    // Verify that start marker point is okay
                    if (msg == null && (ai.getNormalizedX1() == -1 || ai.getNormalizedY1() == -1))
                    {
                        if (ai.getNormalizedX1() != ai.getNormalizedY1())       // both must equal -1
                            msg = String.format("annotation %d. Partially defined start point.", n);
                    }

                    // Verify that end marker point is okay
                    if (msg == null && (ai.getNormalizedX2() == -1 || ai.getNormalizedY2() == -1))
                    {
                        if (ai.getNormalizedX2() != ai.getNormalizedY2())       // both must equal -1
                            msg = String.format("annotation %d. Partially defined end point.", n);
                    }

                    // Another test whether there is an end marker point but not a start marker point
                    if (msg == null && ai.getNormalizedX2() != -1 && ai.getNormalizedX1() == -1)
                        msg = String.format("annotation %d. End point exists without a start point.", n);

                    // if there is a start text offset but no end text offset
                    if (msg == null && ai.getTextStart(appstate.getTextMap()).row != -1 &&
                                       ai.getTextEnd(appstate.getTextMap()).row == -1)
                        msg = String.format("annotation %d. Start text offset exists without an end text offset.", n);

                    // if there is a end text offset but no start text offset
                    if (msg == null && ai.getTextEnd(appstate.getTextMap()).row != -1 &&
                                       ai.getTextStart(appstate.getTextMap()).row == -1)
                        msg = String.format("annotation %d. End text offset exists without a start text offset.", n);

                    if (msg != null)
                    {
                        errorReported = true;
                        bw.write(filename);
                        bw.write(" ");
                        bw.write(msg);
                        bw.write("\n");
                        summary.get(typeError).put(filekey, msg);
                    }

                    if (!errorReported)             // save which of the 7 types of annotations this one is
                    {
                        int type = -1;
                        boolean commentExists = !(ai.getComment() == null || ai.getComment().length() == 0);
                        boolean markersExist = ai.getNormalizedX1() != -1;
                        boolean highlightExists = !ai.getTextStart(appstate.getTextMap()).equals(new TextPoint(-1, -1));

                        if (commentExists)
                        {
                            if (markersExist && highlightExists)
                            {
                                commentMarkersHighlight++;
                                type = typeCommentMarkersHighlight;
                            }
                            else if (markersExist)
                            {
                                commentMarkers++;
                                type = typeCommentMarkers;
                            }
                            else if (highlightExists)
                            {
                                commentHighlight++;
                                type = typeCommentHighlight;
                            } else
                            {
                                commentOnly++;
                                type = typeCommentOnly;
                            }
                        } else if (markersExist && highlightExists)
                        {
                            markersHighlight++;
                            type = typeMarkersHighlight;
                        } else if (markersExist)
                        {
                            markersOnly++;
                            type = typeMarkersOnly;             // should have been spotted earlier
                        } else if (highlightExists)
                        {
                            highlightOnly++;
                            type = typeHighlightOnly;           // should have been spotted earlier
                        }
                        if (type == -1)
                            throw new Exception("Why is type -1");
                        else
                        {
                            if (type == typeMarkersHighlight || type == typeCommentMarkersHighlight)
                                sortList.add(ai);

                            String s = summary.get(type).get(filekey);
                            String delim = ", ";
                            if (s == null)
                            {
                                s = "see annotation";
                                delim = " ";
                            }
                            else if (s.charAt(14) == ' ')
                                s = "see annotations" + s.substring(15);        // pluralize it

                            summary.get(type).put(filekey, String.format("%s%s%d", s, delim, n));
                        }
                    }
                }

                // One last test: if more than two annotations have both markers and a highlight then verify that markers and highlights are in same order.
                int len = sortList.size();
                if (len > 1)
                {
                    sortAnnotations(sortList);
                    boolean okay = true;
                    for (int i = 0; i < len - 1 && okay; i++)         // they are now in order by highlight offset. Verify that they are also in order by markers
                    {
                        AnnotationInfo ai1 = sortList.get(i);
                        AnnotationInfo ai2 = sortList.get(i + 1);
                        okay = ai2.getNormalizedY1() - ai1.getNormalizedY1() > 50
                                || ((ai2.getNormalizedY1() - ai1.getNormalizedY2() <=50) && ai1.getNormalizedX1() < ai2.getNormalizedX2());

                        if (!okay)
                        {
                            Hashtable<String, String> h = summary.get(typeOutOfOrder);
                            String s = h.get(filekey);
                            if (s == null)
                                s = "see annotations";
                            s = String.format("%s %d & %d", s, ai1.getSequenceNumber(), ai2.getSequenceNumber());
                            summary.get(typeOutOfOrder).put(filekey, s);
                        }
                    }
                }
            }
        }

//        if (currentDirname != null && currentFilename != null)
//            appstate.getAnnotater().reloadAnnotations(currentDirname, currentFilename);

        bw.close();

        return summary;
    }


    public ArrayList<Hashtable<String, String>> generateCommentsSummary(String dirname, boolean useCurrentManuscript) throws Exception
    {
        String relativeDirname = null;
        Annotations annotations = new Annotations(appstate);
        appstate.commentsList = new ArrayList<RawComment>();
        ArrayList<Hashtable<String, String>> summary = new ArrayList<Hashtable<String, String>>();

        Regex re = new Regex(dirname, String.format("%s/(.+)$", appstate.getAnnotationsRootDirectory()));
        if (re.find())
            relativeDirname = re.getGroup(1);
        String outfile = dirname + "/annotationSummary.txt";
        BufferedWriter bw = IOAssistance.openWriter(outfile);

        ArrayList<String> allManuscriptDirectories = null;

        if (useCurrentManuscript)
        {
            allManuscriptDirectories = new ArrayList<String>();
            allManuscriptDirectories.add(appstate.getLastManuscript());
        }
        else
            allManuscriptDirectories = appstate.getAllManuscriptDirectories();


        ArrayList<CommentInfo> tags = appstate.getCommentTags();
        for (int i=0; i<tags.size(); i++)
        {
            tags.get(i).sequenceNumber = i;
            summary.add(new Hashtable<String, String>());
        }
        Collections.sort(tags);

        String archetypeDirectory = String.format("%s/%s", appstate.getAnnotationsRootDirectory(), appstate.getLastText());  // e.g. annotations/BhagavadGita

        for (String dn : allManuscriptDirectories)
        {
            File df = new File(String.format("%s/%s", archetypeDirectory, dn));
            if (df.exists())
            {
                File[] ff = df.listFiles();
                String filekey = null;

                for (File f : ff)
                {
                    String filename = f.getAbsolutePath();
                    int n = 0;
                    String lastpart = f.getName();
                    if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
                    {
                        Regex rex = new Regex(filename, "[\\\\/]([^\\\\/]+)\\.xml$");
                        rex.find();
                        filekey = rex.getGroup(1);                  // this is used to associate summary information with this xml file

                        annotations.loadAnnotations(relativeDirname, f.getAbsolutePath());         // these will be normalized
                        ArrayList<AnnotationInfo> ais = annotations.getAnnotations();
                        n = 0;      // annotation number
                        for (AnnotationInfo ai : ais)
                        {
                            if (ai.getComment().length() > 0)
                            {
                                String comment = ai.getComment();
                                appstate.commentsList.add(new RawComment(comment.toLowerCase(), filekey, n));
                                for (CommentInfo ci : tags)
                                {
                                    if (comment.indexOf(ci.description) != -1)
                                    {
                                        String s = summary.get(ci.sequenceNumber).get(filekey);
                                        String delim = ", ";
                                        if (s == null)
                                        {
                                            s = "see annotation";
                                            delim = " ";
                                        }
                                        else if (s.charAt(14) == ' ')
                                            s = "see annotations" + s.substring(15);        // pluralize it

                                        summary.get(ci.sequenceNumber).put(filekey, String.format("%s%s%d", s, delim, n+1));
                                        break;
                                    }
                                }
                            }
                            n++;
                        }
                    }
                }
            }
        }

        Collections.sort(tags, new CommentInfo(""));            // put back into original order

        bw.close();

        return summary;
    }


    public ArrayList<VerseAnnotationCollection> generateDifferencesSummary(String currentDirectory, String backupDirectory, boolean useCurrentManuscript) throws Exception
    {
        if (!new File(backupDirectory).exists())
            return null;
        Annotations annotations = new Annotations(appstate);
        File dir = new File(currentDirectory);
        File[] ff = dir.listFiles();
        String backedUpFileName = null;
        ArrayList<VerseAnnotationCollection> summary = new ArrayList<VerseAnnotationCollection>();
        Annotations backupAnnotations = new Annotations(appstate);

        //String currentDirname = appstate.annotations.getDirname();
        //String currentFilename = appstate.annotations.getFilename();

        for (File f : ff)
        {
            VerseAnnotationCollection vac = null;
            String filename = f.getAbsolutePath();
            String lastpart = f.getName();
            if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
            {
                Regex rex = new Regex(filename, "[\\\\/]([^\\\\/]+)\\.xml$");
                rex.find();
                String backupKey = rex.getGroup(1);
                backedUpFileName = String.format("%s/%s.xml", backupDirectory, backupKey);

                annotations.loadAnnotations(currentDirectory, f.getAbsolutePath());         // these will be normalized
                ArrayList<AnnotationInfo> ais = annotations.getAnnotations();

                File bfn = new File(backedUpFileName);
                if (bfn.exists())
                    backupAnnotations.loadAnnotations(backupDirectory, bfn.getAbsolutePath());
                else
                {
                    int len = ais.size();
                    for (int i=0; i<len; i++)
                    {
                        vac = updatevac(vac, backupKey, i, summary);
                    }
                    continue;
                }

                ArrayList<AnnotationInfo> bais = backupAnnotations.getAnnotations();
                if (bais.size() > ais.size())
                    vac = updatevac(vac, backupKey, -(bais.size()-ais.size()), summary);
                int n = 0;
                for (AnnotationInfo ai : ais)       // try to find each current annotation in the backup annotations
                {
                    boolean found = false;
                    for (AnnotationInfo bai : bais)
                    {
                        if (ai.compare(bai))
                        {
                            found = true;
                            break;
                        }
                    }
                    if (!found)
                        vac = updatevac(vac, backupKey, n, summary);
                    n++;
                }
            }
        }

//        if (currentDirname != null && currentFilename != null)
//            appstate.getAnnotater().reloadAnnotations(currentDirname, currentFilename);

        return summary;
    }

    private VerseAnnotationCollection updatevac(VerseAnnotationCollection vac, String backupKey, int n, ArrayList<VerseAnnotationCollection> summary)
    {
        if (vac == null)
        {
            vac = new VerseAnnotationCollection(backupKey, n);
            summary.add(vac);
        } else
            vac.annotationNumbers.add(n);

        return vac;
    }


    public void searchForText(boolean useCurrentManuscript, String query, ArrayList<VerseAnnotationEx> foundImages) throws Exception
    {
        useCurrentManuscript = true;
        Annotations annotations = new Annotations(appstate);

        foundImages.clear();

        String textFilename = appstate.getLastText();
        String textDirectory = appstate.basicConfig.getTextDirectory();
        ArrayList<String> allManuscriptDirectories = null;

        if (useCurrentManuscript)
        {
            allManuscriptDirectories = new ArrayList<String>();
            allManuscriptDirectories.add(appstate.getLastManuscript());
        }
        else
            allManuscriptDirectories = appstate.getAllManuscriptDirectories();

        String archetypeDirectory = String.format("%s/%s", appstate.getAnnotationsRootDirectory(), appstate.getLastText());  // e.g. annotations/BhagavadGita
        for (String dn : allManuscriptDirectories)
        {
            File df = new File(String.format("%s/%s", archetypeDirectory, dn));
            if (df.exists())
            {
                String encodingTypeFile = df.getAbsolutePath()+"/encodingType";
                String encodingType = null;
                if (new File(encodingTypeFile).exists())
                    encodingType = IOAssistance.readEntireFile(encodingTypeFile);
                else
                    encodingType = appstate.getDefaultEncoding();

                String slp1Filename = String.format("%s/%s.%s", textDirectory, textFilename, encodingType);
                String slp1Text = IOAssistance.readEntireFile(slp1Filename);

                String map = String.format("*:slp1->%s", encodingType);
                tm = new TransformMap(map, ris, fsmh);

                String encoding = appstate.getEncoding();                   // current encoding of digital text
                if (appstate.getSearchWithSLP1() && !encoding.equals("slp1"))    // convert to SLP1 if appropriate
                {
                    // Replace regular expression metacharacters so that they don't get converted
                    query = query.replaceAll("\\\\", "\u0001");
                    query = query.replaceAll("\\[", "\u0002");
                    query = query.replaceAll("\\]", "\u0003");

                    query = WebServices.transformString(query, tm);

                    // Put metacharacters back.
                    query = query.replaceAll("\\[\u0001\\]", "\\\\");            // put escape back
                    query = Utils.removeBraces(query, '[');
                    query = query.replaceAll("\u0002", "[");
                    query = query.replaceAll("\u0003", "]");
                }

                String[] slp1Lines = slp1Text.split("\n");

                eolLength = getEolLength(slp1Text);

                String path = String.format("%s/%s/%s", appstate.basicConfig.getAnnotationsRootDirectory(), appstate.getLastText(), dn);
                File dir = new File(path);
                File[] ff = dir.listFiles();
                for (File f : ff)
                {
                    String filename = f.getAbsolutePath();
                    String lastpart = f.getName();
                    if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
                    {
                        try
                        {
                            annotations.loadAnnotations(path, f.getAbsolutePath());
                            ArrayList<AnnotationInfo> ais = annotations.getAnnotations();
                            int n = 0;
                            for (AnnotationInfo ai : ais)
                            {
                                if (ai.hasSelectedText())
                                    searchRange(dn, n, filename, ai.getTextStart(appstate.getTextMap()),
                                            ai.getTextEnd(appstate.getTextMap()), slp1Lines, query, foundImages);
                                n++;
                            }
                        }
                        catch (Exception ex)
                        {
                            Logger.writeLog(String.format("%s probably has bad XML in it. Please fix, change extension to something other than .xml or remove from directory.", filename));
                        }
                    }
                }
            }
        }
    }

    private void sortAnnotations(ArrayList<AnnotationInfo> a)
    {
        int len = a.size();
        int j = len-1;
        boolean sorted = false;
        while (!sorted)
        {
            sorted = true;
            for (int i = 0; i < j; i++)
            {
                AnnotationInfo ai1 = a.get(i);
                AnnotationInfo ai2 = a.get(i + 1);
                if (ai1.getTextStart(appstate.getTextMap()).row > ai2.getTextStart(appstate.getTextMap()).row ||
                        (ai1.getTextStart(appstate.getTextMap()).row == ai2.getTextStart(appstate.getTextMap()).row && ai1.getTextStart(appstate.getTextMap()).col > ai2.getTextStart(appstate.getTextMap()).col))
                {
                    sorted = false;
                    a.set(i, ai2);
                    a.set(i + 1, ai1);
                }
            }
            j--;
        }
    }

    public boolean haveSelectedText(String dirname) throws Exception
    {
        Annotations annotations = new Annotations(appstate);
        boolean rc = false;
        //String currentDirname = appstate.annotations.getDirname();
        //String currentFilename = appstate.annotations.getFilename();
        File dir = new File(dirname);
        File[] ff = dir.listFiles();
        for (File f : ff )
        {
            String lastpart = f.getName();
            int n = 0;
            if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
            {
                annotations.loadAnnotations(dirname, f.getAbsolutePath());
                ArrayList<AnnotationInfo> ais = annotations.getAnnotations();
                for (AnnotationInfo ai : ais)
                {
                    if (ai.hasSelectedText())
                    {
                        rc = true;
                        break;
                    }
                }
            }
            if (rc)
                break;
        }

//        if (currentDirname != null && currentFilename != null)
//            appstate.annotations.loadAnnotations(currentDirname, currentFilename);

        return rc;
    }

    public boolean reformat(String oldEncoding, String newEncoding, StringBuilder result) {
        String path = null;
        File tempDirectory = null;
        try {
            boolean modified = false;
            this.oldEncoding = oldEncoding;
            this.newEncoding = newEncoding;
            String lastManuscript = appstate.getLastManuscript();

            path = String.format("%s/%s/%s", appstate.basicConfig.getAnnotationsRootDirectory(), appstate.getManuscriptType(), lastManuscript);
            tempDirectory = new File(String.format("temp/%s/%s_%s_%s", appstate.getManuscriptType(), lastManuscript, oldEncoding, Utils.getDateTime()));
            tempDirectory.mkdirs();
            if (appstate.haveSelectedText(path)) // This test is needed in case the manuscript has multiple textTypes
            {
                modified = true;
                if (Utils.copyDirectory(new File(path), tempDirectory)) {
                    adjustAnnotationOffsets(path, appstate.getLastManuscript(), result);
                } else {
                    Utils.displayError("Attempt to backup directory failed. Nothing converted.");
                    return false;
                }
            }
            appstate.saveEncodingFile(path, newEncoding);


            if (modified) {
                Utils.deleteDirectory(tempDirectory);
            }
        } catch (Exception ex) {
            boolean success = false;
            Logger.writeLog(ex);
            if (Utils.deleteDirectory(new File(path))) {
                if (Utils.copyDirectory(tempDirectory, new File(path))) {
                    Utils.deleteDirectory(tempDirectory);
                    success = true;             // i.e. directory restored.
                    Utils.displayError("Something went wrong. However, your annotation directory was successfully restored. See log.txt for details.");
                }
            }
            if (!success) {
                Utils.displayError("Something went wrong and attempt to restore original annotation directory also failed. Hope you have a backup.");
            }

            return false;
        }

        return true;
    }


    public void createVerseIndex(Hashtable<String, ArrayList<VerseAnnotation>> verseIndex, boolean useCurrentManuscript) throws Exception
    {
        Annotations annotations = new Annotations(appstate);
        this.verseIndex = verseIndex;

        verseIndex.clear();

        String textFilename = appstate.getLastText();
        String textDirectory = appstate.basicConfig.getTextDirectory();
        ArrayList<String> allManuscriptDirectories = null;

        if (useCurrentManuscript)
        {
            allManuscriptDirectories = new ArrayList<String>();
            allManuscriptDirectories.add(appstate.getLastManuscript());
        }
        else
            allManuscriptDirectories = appstate.getAllManuscriptDirectories();

        String archetypeDirectory = String.format("%s/%s", appstate.getAnnotationsRootDirectory(), appstate.getLastText());  // e.g. annotations/BhagavadGita
        for (String dn : allManuscriptDirectories)
        {
            File df = new File(String.format("%s/%s", archetypeDirectory, dn));
            if (df.exists())
            {
                String encodingTypeFile = df.getAbsolutePath()+"/encodingType";
                String encodingType = null;
                if (new File(encodingTypeFile).exists())
                    encodingType = IOAssistance.readEntireFile(encodingTypeFile);
                else
                    encodingType = appstate.getDefaultEncoding();

                String slp1Filename = String.format("%s/%s.%s", textDirectory, textFilename, encodingType);
                String slp1Text = IOAssistance.readEntireFile(slp1Filename);

                String[] slp1Lines = slp1Text.split("\n");

                eolLength = getEolLength(slp1Text);

                String path = String.format("%s/%s/%s", appstate.basicConfig.getAnnotationsRootDirectory(), appstate.getLastText(), dn);
                File dir = new File(path);
                File[] ff = dir.listFiles();
                for (File f : ff)
                {
                    String filename = f.getAbsolutePath();
                    String lastpart = f.getName();
                    if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
                    {
                        try
                        {
                            annotations.loadAnnotations(path, f.getAbsolutePath());
                            ArrayList<AnnotationInfo> ais = annotations.getAnnotations();
                            int n = 0;
                            for (AnnotationInfo ai : ais)
                            {
                                if (ai.hasSelectedText())
                                    processRange(dn, n, filename, ai.getTextStart(appstate.getTextMap()), ai.getTextEnd(appstate.getTextMap()), slp1Lines, slp1Text);
                                n++;
                            }
                        }
                        catch (Exception ex)
                        {
                            Logger.writeLog(String.format("%s probably has bad XML in it. Please fix, change extension to something other than .xml or remove from directory.", filename));
                        }
                    }
                }
            }
        }

        Enumeration<String> en = verseIndex.keys();
        ArrayList<String> a = new ArrayList<String>();
        while (en.hasMoreElements())
        {
            a.add(en.nextElement());
        }

        Collections.sort(a);

        try
        {
            BufferedWriter bw = IOAssistance.openWriter("verseIndexAll.txt");
            for (String s : a)
            {
                bw.write(s);
                ArrayList<VerseAnnotation> aa = verseIndex.get(s);

                if (aa.size() > 1)
                    bw.write("*");

                bw.write(" ");

                for (VerseAnnotation f : aa)
                {
                    if (f.manuscriptDirectory != null)
                        bw.write(String.format("%s/", f.manuscriptDirectory));
                    Regex re = new Regex(f.filename, "(\\d*\\.xml)$");
                    if (re.find())
                        bw.write(re.getGroup(1) + ".jpg");
                    else
                        bw.write(f.filename);
                    bw.write(" ");
                }
                bw.write("\n");
            }
            bw.close();
        }
        catch (Exception ex)
        {
        }
    }

    private void adjustAnnotationOffsets(String annotationsPath, String textFilename, StringBuilder result) throws Exception {
        String map = String.format("*:%s->%s", oldEncoding, newEncoding);
        tmOldNew = new TransformMap(map, ris, fsmh);
        map = String.format("*:%s->%s", newEncoding, oldEncoding);
        _tmNewOld = new TransformMap(map, ris, fsmh);       // for round trip
/*
        // tmNewSLP1, _tmSLP1New, tmSLP1Old, _tmOldSLP1 are not needed if brackets are not being removed.
        if (!newEncoding.equals("slp1")) {
            map = String.format("*:%s->slp1", newEncoding);
            tmNewSLP1 = new TransformMap(map, ris, fsmh);

            map = String.format("*:slp1->%s", newEncoding);
            _tmSLP1New = new TransformMap(map, ris, fsmh);
        }

        if (!oldEncoding.equals("slp1")) {
            map = String.format("*:slp1->%s", oldEncoding);
            tmSLP1Old = new TransformMap(map, ris, fsmh);

            map = String.format("*:%s->slp1", oldEncoding);
            _tmOldSLP1 = new TransformMap(map, ris, fsmh);
        }
*/
        setupContexts();                // for roundtrips

        // Now read each annotation .xml file in the directory and convert the text elements from the old encoding to the new encoding
        File dir = new File(annotationsPath);
        File[] ff = dir.listFiles();
        for (File f : ff) // for each .xml file
        {
            String filename = f.getAbsolutePath();
            String lastpart = f.getName();
            if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.') // for each .xml file
            {
                appstate.annotations.loadAnnotations(annotationsPath, f.getAbsolutePath());
                ArrayList<AnnotationInfo> ais = appstate.annotations.getAnnotations();
                boolean selectedText = false;
                int annotationNumber = 0;
                for (AnnotationInfo ai : ais) {         // for each annotation
                    annotationNumber++;
                    boolean isCurrentTab = ai.getTabNumber() == appstate.getCurrentTabNumber();
                    boolean hasOldEncoding = ai.getEncoding().equals(oldEncoding);
                    if (isCurrentTab && hasOldEncoding) {
                        ai.setEncoding(newEncoding);
                    }
                    if (ai.hasSelectedText() && hasOldEncoding && isCurrentTab) {
                        selectedText = true;
                        ArrayList<TextBlock> textBlocks = ai.getTextMarkers();
                        int len = textBlocks.size();
                        for (int i = 0; i < len; i++) {    // for each text block in the annotation
                            TextBlock newtb = null;
                            TextBlock textBlock = textBlocks.get(i);
                            getTexts(textBlock.filename);       // will set oldLines and newLines, reading from files if necessary.
                            forwardContext.oldLines = oldLines; // for transforming from old transcoding to new transcoding
                            forwardContext.newLines = newLines;
                            returnContext.oldLines = newLines;  // for the return trip to verify that no information lost in the forward transformation
                            returnContext.newLines = oldLines;
                            newtb = adjustColumns(textBlock.filename, textBlock, forwardContext);       // update start and end cols
                            TextBlock oldtb = adjustColumns(textBlock.filename, newtb, returnContext);  // return trip to verify that original start and end cols are recomputed.

                            if (!textBlock.isEqualTo(oldtb)) // verify that a round trip produced same result.
                            {
                                result.append(String.format("%s: annotation %d\n", filename, annotationNumber));
                                result.append(getAnnotationText(textBlock, oldLines));
                                result.append("\nadjusted to\n");
                                result.append(getAnnotationText(newtb, newLines));
                                result.append("\n\n");
                                throw new Exception(String.format("%s, annotation %d before=%d.%d, after roundtrip=%d.%d",
                                        filename, annotationNumber,
                                        textBlock.startRow, textBlock.startCol,
                                        newtb.startRow, newtb.startCol));
                            }

                            textBlocks.set(i, newtb);           // update text block
                        }
                        ai.setEncoding(newEncoding);
                    }
                }

                if (selectedText) {         // if any of the annotations in this file got transformed

                    appstate.annotations.toXml(appstate.annotations.getResolvedWidth());
                }
            }
        }
    }


    // For building an error message string
    private String getAnnotationText(TextBlock tb, String[] lines) {
        StringBuilder sbb = new StringBuilder();

        for (int i=tb.startRow; i<=tb.endRow; i++) {
            String line = lines[i];
            if (i == tb.startRow) {
                line = line.substring(tb.startCol, tb.endCol+1);
            } else if (i == tb.endRow) {
                line = line.substring(0, tb.endCol+1);
            }
            sbb.append(line);
        }
        
        return sbb.toString();
    }


    // Set oldLine and newLines, reading from file the first time.
    private void getTexts(String textFilename) throws Exception {
        if (textLines.get(textFilename) != null) {
            oldLines = textLines.get(textFilename)[0];
            newLines = textLines.get(textFilename)[1];
        } else {
            String textDirectory = appstate.basicConfig.getTextDirectory();
            String oldFilename = String.format("%s/%s.%s", textDirectory, textFilename, oldEncoding);
            oldText = IOAssistance.readEntireFile(oldFilename) + "  \n  \n";              // add an extra line in case some selection goes past last \n
            String newFilename = String.format("%s/%s.%s", textDirectory, textFilename, newEncoding);
            newText = IOAssistance.readEntireFile(newFilename) + "  \n  \n";              // add an extra line in case some selection goes past last \n

            oldLines = oldText.split("\n");
            newLines = newText.split("\n");

            if (oldLines.length != newLines.length) {
                throw new Exception(String.format("%s has %d lines but %s has %d lines. They should be the same!"));
            }
            textLines.put(textFilename, new String[][] {oldLines, newLines});
        }
    }

    private void setupContexts()
    {
        forwardContext = new ConversionContext();
        forwardContext.oldEncoding = oldEncoding;
        forwardContext.newEncoding = newEncoding;
        forwardContext.oldText = oldText;
        forwardContext.oldLines = oldLines;
        forwardContext.newText = newText;
        forwardContext.newLines = newLines;
        forwardContext.tmOldNew = tmOldNew;
        forwardContext.tmNewSLP1 = tmNewSLP1;
        forwardContext.tmSLP1Old = tmSLP1Old;

        returnContext = new ConversionContext();
        returnContext.oldEncoding = newEncoding;
        returnContext.newEncoding = oldEncoding;
        returnContext.oldText = newText;
        returnContext.oldLines = newLines;
        returnContext.newText = oldText;
        returnContext.newLines = oldLines;
        returnContext.tmOldNew = _tmNewOld;
        returnContext.tmNewSLP1 = _tmOldSLP1;
        returnContext.tmSLP1Old = _tmSLP1New;
    }

    
    private TextBlock adjustColumns(String filename, TextBlock tb, ConversionContext cc) throws Exception {
        String oldLine = insertMarker(cc.oldLines[tb.startRow], tb.startCol);
        TextBlock newtb = new TextBlock(tb);
        newtb.startCol = findMarker(WebServices.transformString(oldLine, cc.tmOldNew));

        oldLine = insertMarker(cc.oldLines[tb.endRow], tb.endCol);
        newtb.endCol = findMarker(WebServices.transformString(oldLine, cc.tmOldNew));

        return newtb;
    }


    private String insertMarker (String line, int col) {
        sb.setLength(0);
        int len = line.length();
        for (int i=0; i<len; i++) {
            if (i == col) {
                sb.append((char)0x04);
            }
            sb.append(line.charAt(i));
        }
        return sb.toString();
    }

    private int findMarker (String line) {
        int len = line.length();
        for (int i=0; i<len; i++) {
            if (line.charAt(i) == 0x04) {
                return i;
            }
        }
        return -1;
    }

/*
    private String removePassthroughs(String text, ConversionContext cc) throws Exception
    {
        if (cc.tmNewSLP1 != null)
            text = WebServices.transformString(text, cc.tmNewSLP1);            // convert target encoding to SLP1 and get rid of any bracketed characters

        text = cleanupBrackets(text);

        if (cc.tmSLP1Old !=  null)
            text = WebServices.transformString(text, cc.tmSLP1Old);   // convert it to the oldEncoding

        return text;
    }


    private String cleanupBrackets(String text)
    {
        sb.setLength(0);
        // replace any bracketed characters with blanks
        int len = text.length();
        for (int i=0; i<len; i++)
        {
            if (text.charAt(i) == '[')
            {
                for (int j=i; j<len; j++)
                {
                    if (text.charAt(j) != ']')
                        i++;
                    else
                        break;
                }

                sb.append(' ');
            }
            else
                sb.append(text.charAt(i));
        }

        return sb.toString();
    }
*/
    // This is not general. It assumes that the verse number is at the beginning of each line and is 8 digits long
    private void processRange(String manuscriptDirectory, int annotationNumber, String filename, TextPoint textStart, TextPoint textEnd, String[] slp1Lines, String slp1Text)
    {
        // TODO
/*
        int[] lineIdxStart = new int[1];
        int[] lineIdxEnd = new int[1];
        int start = Utils.getLineNumber(slp1Text, textStart, lineIdxStart);
        int end = Utils.getLineNumber(slp1Text, textEnd, lineIdxEnd);
        if (lineIdxEnd[0] == textEnd)
            end--;                      // cursor is at beginning of line

        Regex re = new Regex("", "^(\\d{8})");
        for (int i=start; i<=end; i++)                  // for each line number that is part of the selection
        {
            String line = slp1Lines[i];
            re.reset(line);
            if (re.find())                              // get the verse number (sans the "a" or "c")
            {
                String vn = re.getGroup(1);
                ArrayList<VerseAnnotation> images = verseIndex.get(vn);
                if (images == null)
                {
                    images = new ArrayList<VerseAnnotation>();
                    verseIndex.put(vn, images);
                }

                VerseAnnotation verseAnnotation = null;
                for (VerseAnnotation va : images)
                {
                    if (va.filename.equals(filename))
                    {
                        
                        verseAnnotation = va;
                        break;
                    }
                }
                if (verseAnnotation == null)             // is this verse mapped to more than one image? e.g., part of verse is on one page and part on another
                {
                    verseAnnotation = new VerseAnnotation(manuscriptDirectory, filename, annotationNumber);
                    images.add(verseAnnotation);
                }

                int highlightWidth = textEnd - textStart;
                if (verseAnnotation.highlightWidth < highlightWidth)
                    verseAnnotation.highlightWidth = highlightWidth;
            }
        }
 */
    }

    private void searchRange(String manuscriptDirectory, int annotationNumber, String filename, TextPoint textStart, TextPoint textEnd,
                                String[] text, String query, ArrayList<VerseAnnotationEx> foundImages) throws Exception {
        for (int i=textStart.row; i<=textEnd.row; i++) {
            Regex re = new Regex(text[i], query);

            if (re.find()) {
                for (VerseAnnotationEx va : foundImages)
                {
                    if (va.manuscriptDirectory.equals(manuscriptDirectory) && va.filename.equals(filename))
                    {
                        va.annotationNumbers.add(annotationNumber);
                        return;
                    }
                }

                foundImages.add(new VerseAnnotationEx(manuscriptDirectory, filename, annotationNumber));
            }
        }
    }

    int getEolLength(String text) throws Exception
    {
        int len = 0;

        for (int i = 0; i < text.length(); i++)
        {
            if (text.charAt(i) <= ' ')
            {
                if (text.charAt(i + 1) <= ' ')
                    len = 2;
                else
                    len = 1;
                break;
            }
        }

        if (len != 1)
            throw new Exception("eolLength != 1");

        return len;
    }

    public void fixupShift()
    {
// TODO
/*
        String textfilename = "C:/Users/ralph/Documents/PeterScharf/Manuscripts/DigitalTexts/BhG.roman";
        String text = null;
        try
        {
            text = IOAssistance.readEntireFile(textfilename);
        }
        catch (Exception ex)
        {

        }
        Annotations annotations = new Annotations(appstate);
        String path = "C:/Users/ralph/Documents/PeterScharf/SanskritLibrary/SITA/annotations/BagavadgItA/UPenn/mscoll390_item492_wk1";
        File dir = new File(path);
        File[] ff = dir.listFiles();
        for (File f : ff)
        {
            String filename = f.getAbsolutePath();
            String lastpart = f.getName();
            if (lastpart.toLowerCase().endsWith(".xml") && lastpart.charAt(0) != '.')        // for each .xml file
            {
                Regex re = new Regex(filename, "00(\\d\\d)\\.xml");
                re.find();
                int pn = Integer.parseInt(re.getGroup(1));
                if (pn <=15)
                    continue;
                try
                {
                    annotations.loadAnnotations(path, f.getAbsolutePath());
                    ArrayList<AnnotationInfo> ais = annotations.getAnnotations();
                    int n = 0;
                    boolean hasSelectedText = false;
                    for (AnnotationInfo ai : ais)
                    {
                        if (ai.hasSelectedText())
                        {
                            hasSelectedText = true;
                            int diff = computeAdjustment(text, ai.getTextStart());
                            ai.setTextStart(ai.getTextStart()-diff);
                            diff = computeAdjustment(text, ai.getTextEnd());
                            ai.setTextEnd(ai.getTextEnd()-diff);
                        }
                        n++;
                    }
                    if (hasSelectedText)
                        annotations.toXml(annotations.getResolvedWidth());

                } catch (Exception ex)
                {
                    Logger.writeLog(String.format("%s probably has bad XML in it. Please fix, change extension to something other than .xml or remove from directory.", filename));
                }
            }
        }
*/
    }

    private int computeAdjustment(String text, int offset)
    {
        int oldOffset = 0;
        int newOffset = 0;

        while (oldOffset != offset)
        {
            char ch = text.charAt(newOffset);
            if (ch==';')
            {
                oldOffset+=2;
            }
            oldOffset++;
            newOffset++;
        }

        return oldOffset - newOffset;
    }
}
