package org.sanskritlibrary;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class FSMEntry implements Comparator
{
	public String id;
    public String key;
    public String edge;
    public ArrayList<String> edges;
    public String ligature;
    public String map;
    public String target;
    public ArrayList<FSMEntry> targets;     // for ligatures
	public String klass;
	public MatchItem[] extension;
	public MatchItem[] context;
    public boolean dependant = false;
    public boolean zeroWidth = false;
    public FSMEntry mappedTo;
    public boolean removed = false;   // used for default entry
    public String comment;
    public boolean swap = false;      // if this character should be swapped with following character
    public boolean ignore = false;    // for instance, Cardona's file used a 1f character that did not show up in PDF but would cause problems if it appeared before a vowel sign

	public FSMEntry()
	{		 
	}

     public FSMEntry(FSMEntry fe)
     {
         this.id = fe.id;
         this.key = fe.key;
         this.edge = fe.edge;
         this.map = fe.map;
         this.ligature = fe.ligature;
         this.target = fe.target;
         this.klass = fe.klass;
         this.extension = fe.extension;
         this.context = fe.context;
         this.dependant = fe.dependant;
         this.zeroWidth = fe.zeroWidth;
         this.comment = fe.comment;
         this.swap = fe.swap;
     }

    public FSMEntry(String id, String edge, String map, String ligature, MatchItem[] extension, MatchItem[] context, String target,
				String klass, boolean dependant, boolean zeroWidth, String key, String comment, boolean swap) throws Exception
	 {
		  if (id == null)
			  throw new Exception("id cannot be null in FSMEntry constructor");
		  else
			  this.id = id;
		  
        this.edge = edge;
        this.map = map;
        this.ligature = ligature;
		this.extension = extension;
		this.context = context;
        this.target = target;
		this.klass = klass;
        this.dependant = dependant;
        this.zeroWidth = zeroWidth;
        this.key = key;
        this.comment = comment;
        this.swap = swap;
    }

    public void addLigatureComponent(FSMEntry fe)
    {
        if (targets == null)
            targets = new ArrayList<FSMEntry>();

        targets.add(fe);
    }

    public void convertEdge()
    {
        if (edge.indexOf("\\u") != -1)
            edge = fixup(edge);     // convert all \\uNNNN to char
    }

    public void convertTarget()
    {
        if (target != null)
        {
            if (target.indexOf("\\u") != -1)
                target = fixup(target);     // convert all \\uNNNN to char
        }
        else
            target = "";
    }

    private static String fixup(String s)
	{
		StringBuilder sb = new StringBuilder();
		int i = 0;
		while (i < s.length())
		{
			Pattern p = Pattern.compile("^[\\\\]u([0-9a-fA-F]{4})");
			Matcher m = p.matcher(s.substring(i));
			if (m.find())
			{
				String hex = m.group(1);
				sb.append((char) Integer.parseInt(hex, 16));
				i += 6;
			} else
			{
				sb.append(s.charAt(i));
				i++;
			}
		}

		return sb.toString();
	}

    static Hashtable<Character, ArrayList<FSMEntry>> getCharOptions(FSMEntry[] fsmEntry) throws Exception
    {
       Hashtable<Character, ArrayList<FSMEntry>> charOptions = new Hashtable<Character, ArrayList<FSMEntry>>();
        int linenum = 0;
        for (FSMEntry fe : fsmEntry)
		{
            if (fe.id.equals("272"))
                fe.id="272";
			linenum++;

			ArrayList<FSMEntry> list = null;

            if (fe.edge != null)
            {
                if (fe.edge.length() == 0) // should not happen
                    throw new Exception("Edge is null");

                if (!charOptions.containsKey(fe.edge.charAt(0)))
                {
                    list = new ArrayList<FSMEntry>();
                    charOptions.put(fe.edge.charAt(0), list);
                }
                else
                    list = charOptions.get(fe.edge.charAt(0));

                list.add(fe);
            }
            else
            {
                for (String s : fe.edges)
                {
                    char ch = s.charAt(0);
                    if (!charOptions.containsKey(ch))
                    {
                        list = new ArrayList<FSMEntry>();
                        charOptions.put(ch, list);
                    }
                    else
                        list = charOptions.get(ch);

                    boolean duplicate = false;
                    for (FSMEntry test : list)
                    {
                        if (test == fe)
                        {
                            duplicate = true;
                            break;
                        }
                    }

                    if (!duplicate)
                        list.add(fe);
                }   
            }
        }
        return charOptions;
    }

    static Hashtable<String, FSMEntry> resolveKeys(FSMEntry[] fsmEntry) throws Exception
    {
        int iterations = 0;
        Hashtable<String, FSMEntry> h = new Hashtable<String, FSMEntry>();      // h maps a key to its FSMEntry
        boolean done = false;

        if (fsmEntry.length == 0)
            return h;               // no entries

        // add user-defined keys
        for (FSMEntry fe : fsmEntry)
        {
            if (fe.key != null)
                h.put(fe.key, fe);
        }

        while (!done)
        {
            iterations++;
            if (iterations > fsmEntry.length)
                break;

            done = true;
            for (FSMEntry fe : fsmEntry)
            {
                if (fe.key == null)
                {
                    if (iterations == fsmEntry.length)
                        iterations = fsmEntry.length;
                    if (fe.map != null)
                    {
                        FSMEntry _mappedTo = h.get(fe.map);
                        if (_mappedTo != null)
                        {
                                fe.map = null;      // this forces key to be set below
                                fe.mappedTo = _mappedTo;
                        }
                    }
                    else if (fe.ligature != null)
                    {
                        fe.key = String.format("%s.%s", fe.klass, fe.ligature);
                        h.put(fe.key, fe);
                    }
                    else if (fe.edge != null)
                        processEdge(fe, h, fsmEntry);
                    done = false;   // do another iteration until no key is null.
                }
            }
        }

        if (done)
        {
            // make one more pass to make sure all ligature edges are resolved
            for (FSMEntry fe : fsmEntry)
            {
                if (edgeNeedsResolving(fe))
                {
                    processEdge(fe, h, fsmEntry);
                    if (edgeNeedsResolving(fe))
                    {
                        done = false;
                        break;
                    }
                }
            }

        }
        if (!done)
        {
            System.out.println("The following entries could not be resolved");
            for (FSMEntry fe : fsmEntry)
            {
                if (fe.key == null)
                {
                    System.out.format("id=%s ", fe.id);
                    if (fe.map != null)
                        System.out.format("map %s not found\n", fe.map);
                    else if (fe.edge != null && fe.edge.charAt(0) == '%')
                        System.out.format("edge %s not resolved\n", fe.edge);
                    else
                        System.out.println("unknown problem");
                }
                else if (edgeNeedsResolving(fe))
                    System.out.format("Edge not resolved: %s\n", fe.edge);
            }
            throw new Exception("Could not resolve entries");
        }

        System.out.format("Iterations required: %d\n", iterations);
        // check that keys are unique
        boolean unique = true;
        for (FSMEntry fe : fsmEntry)
        {
           if (h.get(fe.key) != fe && !fe.key.equals("."))
           {
               System.out.format("id=%s and id=%s have the same key (%s)!\n", h.get(fe.key).id, fe.id, fe.key);
               unique = false;
           }
        }

        if (!unique)
            throw new Exception("duplicate keys found.");


        return h;
    }


    private static boolean edgeNeedsResolving(FSMEntry fe)
    {
        return fe.edge != null && fe.edge.charAt(0)=='%' && fe.edge.length() > 1;
    }


    private static void processEdge(FSMEntry fe, Hashtable<String, FSMEntry> h, FSMEntry[] fsmEntry)
    {
        if (fe.edge.charAt(0) != '%' || fe.edge.length() == 1)   //!! Hack, check if length is 1.
        {
            fe.key = String.format("%s.%s", fe.klass, fe.target);  // edge is a literal
            h.put(fe.key, fe);
        }
        else if (fe.edges == null)      // edge is a key string (%xxx.y...). edges will contain all possible combinations
        {
            ArrayList<String> _edges = resolve(fe.edge, fsmEntry, h);
            if (_edges != null)
            {
                if (fe.map == null)
                {
                    fe.key = String.format("%s.%s", fe.klass, fe.target);
                } else
                    fe.key = ".";       // (mapped items do not need keys) so that this entry is not processed again
                fe.edges = _edges;
                fe.edge = null;
                h.put(fe.key, fe);
            }
        }
    }


    private static ArrayList<String> resolve(String s, FSMEntry[] fsmEntry, Hashtable<String, FSMEntry> h)
    {
        StringBuilder sb = new StringBuilder();
        ArrayList<String> a = new ArrayList<String>();
        ArrayList<ArrayList<String>> work = new ArrayList<ArrayList<String>>();
        Regex re = new Regex(s, "([^%]+)");
        while (re.find())
        {
            String key = re.getGroup(1);
            ArrayList<String> aa = getEdges(key, fsmEntry, h);
            if (aa == null)
                return null;
            else 
                work.add(aa);
        }

        if (work.size() > 0)
        {
            int len = work.size();
            int[] end = new int[len];
            int[] current = new int[len];
            current[len-1] = -1;
            for (int i=0; i<len; i++)
            {
                end[i] = work.get(i).size();
            }
            while (nextPermutation(current, end))
            {
                sb.setLength(0);
                for (int i=0; i<len; i++)
                {
                    sb.append(work.get(i).get(current[i]));
                }
                a.add(sb.toString());
            }
        }
        else
            a = work.get(0);

        return a;
    }

    public static boolean nextPermutation(int[] current, int[] end)
    {
        boolean rc = true;
        for (int i=current.length-1; i>=0 && rc; i--)
        {
            current[i]++;
            if (current[i] == end[i])
            {
                if (i == 0)
                    break;
                else
                    current[i] = 0;
            }
            else
                return true;
        }

        return false;
    }


    private static ArrayList<String> getEdges(String key, FSMEntry[] fsmEntry, Hashtable<String, FSMEntry> h)
    {
        FSMEntry fe = h.get(key);
        if (fe == null)
            return null;

        ArrayList<String> edges = new ArrayList<String>();
        ArrayList<FSMEntry> a = new ArrayList<FSMEntry>();

        // now go and get all edges that map to this edge.
        for (FSMEntry fse : fsmEntry)
        {
            if (fse.map != null && fse.mappedTo == null)
                return null;        // not ready for this yet.
            else if (fse == fe || fse.mappedTo == fe)
                a.add(fse);
        }

        // now get edge characters from each FSMEnty
        for (FSMEntry fse : a)
        {
            if (fse.edges != null)
            {
                for (String edge : fse.edges)
                {
                    edges.add(edge);
                }
            }
            else if (fse.edge.charAt(0) != '%')
                edges.add(fse.edge);
            else
                return null;
        }
        
        return edges;
    }

    public int compare(Object o1, Object o2)
    {
        FSMEntry fe1 = (FSMEntry)o1;
        FSMEntry fe2 = (FSMEntry)o2;
        if (!fe1.edge.equals(fe2.edge))
            return fe1.edge.compareToIgnoreCase(fe2.edge);
        else if (fe1.target.length() != fe2.target.length())
            return fe1.target.length() - fe2.target.length();
        else
            return 0;       // this should not happen!
    }
}
