package org.sanskritlibrary;

import java.util.ArrayList;
import java.util.Hashtable;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class Member
{
    public String id;
    public String init;
    public String map;
    public Member mappedTo;
    public boolean required;
    public ArrayList<FSMEntry> fsmEntry = new ArrayList<FSMEntry>();
    public ActionType[] actions;
    public Buffer buffer;
    private Buffer initbuffer;  // buffer that needs to be initialized when something is added to this member.
    private boolean error = false;
    private boolean entryAdded = false;
    StringBuilder sb = new StringBuilder();
    int count;

    public Member(Buffer buffer, XML doc) throws Exception
    {
        id = doc.getAttribute("id");
        init = doc.getOptionalAttribute("init", null);
        map = doc.getOptionalAttribute("map", null);
        required = doc.getOptionalAttribute("required", "n").toLowerCase().equals("y");

        String actionValue = doc.getOptionalAttribute("action", null);
        if (actionValue != null)
            this.actions = parseActions(actionValue);
        this.buffer = buffer;
    }

    public Member(Buffer buffer, Attributes atts) throws SAXException
    {
        id = atts.getValue("id");
        init = atts.getValue("init");
        map = atts.getValue("map");
        String value = atts.getValue("required");
        if (value == null)
            required = false;
        else
            required = value.toLowerCase().equals("y");

        String actionValue = atts.getValue("action");
        if (actionValue != null)
        {
            try
            {
                this.actions = parseActions(actionValue);
            }
            catch (Exception ex)
            {
                throw new SAXException(ex.getMessage());
            }
        }
        this.buffer = buffer;
    }

    public void addFSMEntry(FSMEntry fe) throws Exception
    {
        entryAdded = true;

        if (fe.target.length() == 0)
            return;
        
        if (map != null && mappedTo == null)
        {
            mappedTo = buffer.buffers.b().getMember(map);
            if (mappedTo == null)
                throw new Exception(String.format("Member %s does not exists", map));
            map = null;
        }

        Member targetMember = this;
        if (mappedTo != null)
            targetMember = mappedTo;
        
        if (!fe.dependant)
            buffer.buffers.b().setBaseFound(true);
        else if (!buffer.buffers.b().getBaseFound())
            buffer.buffers.b().setError(true);

        if (fe.target.length() > 0)
        {
            if (fe.zeroWidth && targetMember.fsmEntry.contains(fe))
                System.out.format("Warning duplicate %s not transcoded\n", TextFormatter.charToSlashU(fe.edge));
            else
            {
                count++;
                targetMember.fsmEntry.add(fe);
            }
        }
    }

    public boolean removeFSMEntry(FSMEntry fe)
    {
        for (FSMEntry fse : fsmEntry)
        {
            if (fse == fe)
            {
                fsmEntry.remove(fe);
                count = fsmEntry.size();
                return true;
            }
        }

        return false;
    }

    public int getEntryCount()
    {
        int cnt = 0;
        for (FSMEntry fe : fsmEntry)
        {
            if (!fe.removed)
                cnt++;
        }
        return cnt;
    }

    public void clearEntries()
    {
        count = 0;
        if (buffer.initfse != null)
            buffer.initfse.removed = false;
        fsmEntry.clear();
    }

    public String makeString()
    {
        sb.setLength(0);
        for (FSMEntry fe : fsmEntry)
        {
            if (!fe.removed)
            {
                if (fe.mappedTo != null)
                    sb.append(fe.mappedTo.target);
                else
                    sb.append(fe.target);
            }
        }

        return sb.toString();
    }

    public void register(Hashtable<String, Member> h)
    {
        h.put(id, this);
    }

    public static ActionType[] parseActions(String action) throws Exception
    {
        String[] f = action.split(",");
        ActionType[] a = new ActionType[f.length];
        int i = 0;
        for (String s : f)
        {
            if (s.equals("!"))
                a[i++] = ActionType.APPEND_BUFFER;
            else if (s.equals("!!"))
                a[i++] = ActionType.APPEND_BUFFER_2;
            else if (s.equals("?"))
                a[i++] = ActionType.APPEND_BUFFER_CONDITIONAL;
            else if (s.equals("|"))
                a[i++] = ActionType.SET_BUFFER_READY;
            else if (s.equals("-"))
                a[i++] = ActionType.CLEAR_BUFFER;
            else if (s.equals("x"))
                a[i++] = ActionType.ERROR;
            else if (s.equals("^"))
                a[i++] = ActionType.INITIALIZE;
            else
                throw new Exception(String.format("Unknown command: %s", s));
        }

        return a;
    }

    void performActions(FSMEntry fe) throws Exception
    {
        entryAdded = false; // set true by addFSMEntry

        for (ActionType at : actions)
        {
            switch (at)
            {
                case APPEND_BUFFER:     // !
                    buffer.buffers.b().appendBuffers(false);
                    addFSMEntry(fe);
                    break;
                case APPEND_BUFFER_2:   // !!
                    buffer.buffers.b().appendBuffers(false);
                    addFSMEntry(fe);
                    buffer.buffers.b().appendBuffers(false);
                    break;
                case APPEND_BUFFER_CONDITIONAL:     // ?
                    buffer.buffers.b().appendBuffers(true);
                    addFSMEntry(fe);
                    break;
                case SET_BUFFER_READY:      // |
                    buffer.buffers.b().setBufferReady(true);
                    break;
                case CLEAR_BUFFER:      // -
                    buffer.clearDefaultMember();
                    buffer.buffers.b().setBufferReady(false);
                    break;
                case INITIALIZE:        // ^
                    if (initbuffer == null)
                        initbuffer = buffer.buffers.b().getBuffer(init);
                    if (initbuffer == null)
                        throw new Exception(String.format("init buffer %s not found", init));
                    initbuffer.initialize();
                    break;
                case ERROR:     // x just like APPEND_BUFFER_2 except that error is printed
                    buffer.buffers.b().setCheckForErrors(false);        // turn off normal error processing and add error at end
                    buffer.buffers.b().appendBuffers(false);
                    addFSMEntry(fe);
                    buffer.buffers.b().appendBuffers(false);
                    buffer.buffers.b().setCheckForErrors(true);
                    buffer.buffers.b().setErrorMark();
                    break;
            }
        }

        if (!entryAdded)
            addFSMEntry(fe);
    }
}
