package org.sanskritlibrary;

public class TextFormatter
{
	// chars to UTF-8, e.g., devanagari a (0x0905 ->%e0%a4%85)
	// see http://en.wikipedia.org/wiki/UTF-8
	public static String charToEscapedUTF8(String s)
	{
		s = Regex.replace(s, "\r\n", "\n");
		StringBuilder utftext = new StringBuilder();

		for (int n = 0; n < s.length(); n++)
		{
			int c = s.charAt(n);

			if (c < 0x80)
			{
				utftext.append((char) c);
			} else if ((c > 127) && (c < 0x800))
			{
				utftext.append("%");
				utftext.append(toHexString((c >> 6) | 192));
				utftext.append("%");
				utftext.append(toHexString((c & 63) | 128));
			} else
			{
				utftext.append("%");
				utftext.append(toHexString((c >> 12) | 224));
				utftext.append("%");
				utftext.append(toHexString(((c >> 6) & 63) | 128));
				utftext.append("%");
				utftext.append(toHexString((c & 63) | 128));
			}
		}

		return utftext.toString();
	}

	public static String escapedUTF8ToChar(String s)
	{
		return null;		// won't implement this until it is needed
	}

	public static String charToEscapedUTF8(char ch)
	{
		return charToEscapedUTF8(new String(new char[]{ch}));
	}

	public static String toHexString(int c)
	{
		String s = Integer.toHexString(c);
		if (s.length() % 2 == 1)
		{
			s = "0" + s;
		}
		return s;
	}

	public static String toHexString(int c, int len)
	{
		String s = Integer.toHexString(c);
        int diff = s.length() - len;
        if (diff == 0)
            return s;
        else if (diff < 0)   // prepend zeros
        {
            switch (diff)
            {
                case -1: return "0" + s;
                case -2: return "00" + s;
                case -3: return "000" + s;
                case -4: return "000" + s;
                case -5: return "000" + s;
                case -6: return "000" + s;
                case -7: return "000" + s;
                default: return "??" + s;
            }
        }
        else        // remove zeroes from the beginning
            return s.substring(diff);
	}


    public static String toHexString(String s)
    {
        return toHexString(Integer.parseInt(s));
    }


    public static String toHexString(String s, int len)
    {
        return toHexString(Integer.parseInt(s), len);
    }




	// e.g., transforms "\\u0905" to char devanagari a
	static public String slashUToChar(String s)
	{
		StringBuilder sb = new StringBuilder();
		Regex re = new Regex(s, "([^\\\\]*)(?:\\\\u([\\da-fA-F]+))?");
		while (re.find())
		{
			sb.append(re.getGroup(1));      // normal text (i.e. not preceded by slash u)
			String cp = re.getGroup(2);     // slash u character
			if (cp != null && cp.length() > 0)
				sb.append((char)Integer.parseInt(cp, 16));
		}
		return sb.toString();
	}

	static public String charToSlashU(String s)
	{
		StringBuilder sb = new StringBuilder();
		for (int i=0; i<s.length(); i++)
		{
			char ch = s.charAt(i);
			sb.append(String.format("\\u%s", toHexString((int)ch, 4)));
		}

		return sb.toString();
	}

    static public String charToSlashU(char ch)
    {
        return String.format("\\u%s", toHexString((int)ch, 4));
    }

	static public String charToSlashUDecimal(String s)
	{
		StringBuilder sb = new StringBuilder();
		for (int i=0; i<s.length(); i++)
		{
			char ch = s.charAt(i);
			sb.append(charToSlashUDecimal(ch));
		}

		return sb.toString();
	}

	static public String charToSlashUDecimal(char ch)
	{
		return String.format("\\u%d", (int)ch);
	}

	static public String entityToChar(String s) throws Exception
	{
		StringBuilder sb = new StringBuilder();
		Regex re = new Regex(s, "&#x([0-9A-Fa-f]+);");
		while (re.find())
		{
			sb.append((char)Integer.parseInt(re.getGroup(1), 16));
		}

		if (sb.length() == 0)
			throw new Exception("Input not in form &#x...;");
		
		return sb.toString();
	}

    // The string s may contain surrogates.
	static public String charToEntity(String s)
	{
		StringBuilder sb = new StringBuilder();

		for (int i=0; i<s.length(); i++)
		{
            int ch1 = s.charAt(i);
            if (ch1 >= 0xD800) {
                int ch2 = s.charAt(i+1);
                i = i + 1;   // they come in pairs
                int cp = 0x10000;
                cp += (ch1 - 0xD800) * 0x400;
                cp += (ch2 - 0xDC00);
                sb.append(String.format("&#x%s;", toHexString(cp)));
            } else if (ch1 >= 128) {
                sb.append(String.format("&#x%s;", toHexString(ch1)));
            }
            else {
                sb.append((char)ch1);
            }
		}

		return sb.toString();
	}


    // s must be a sequence of \\u numbers. Some \\u maybe be > 16 bit, if so they are broken into two surrogates, e.g.
    // if s == "\\u1d160", this returns "\\ud834\\udd60"
    static public String rewriteWithSurrogates(String s) {
        StringBuilder sb = new StringBuilder();
		Regex re = new Regex(s, "\\\\u([\\da-fA-F]+)");
		while (re.find())
		{
			String cp = re.getGroup(1);     // slash u character
            if (cp != null) {
                int len = cp.length();
                if (len == 4)
                    sb.append(String.format("\\\\u%s", cp));
                else {
                    int k = Integer.parseInt(cp, 16);
                    k = k - 0x10000;
                    int left = 0xD800 + (k / 0x400);
                    int right = 0xDC00 + (k % 0x400);
                    sb.append(String.format("\\u%s", toHexString(left)));
                    sb.append(String.format("\\u%s", toHexString(right)));
                }
            }
		}

        return sb.toString();
    }

    // This assumes that s contains only codepoints that begin with "\\u". Returns a regular character string with 16 bit characters, e.g.,
    // if s == "\\u1d160", this returns a string result where result.charAt(0) == 0xd834 and result.charAt(1) == 0xdd60.
    public static String convertToLowerSurrogates(String s) {
        StringBuilder sb = new StringBuilder();
		Regex re = new Regex(s, "\\\\u([\\da-fA-F]+)");
		while (re.find())
		{
			String cp = re.getGroup(1);     // slash u character
			if (cp != null) {
                int k = Integer.parseInt(cp, 16);
                int len = cp.length();
                if (len == 4) {
                    sb.append((char)k);
                }
                else if (len > 4) {
                    k -= 0x10000;
                    int left = k/0x400 + 0xD800;
                    int right = (k%0x400) + 0xDC00;
                    sb.append((char)left);
                    sb.append((char)right);
                }
            }
		}

        return sb.toString();
//      int n = 0x10000 + (0xD834 - 0xD800) * 0x400 + (0xDD60 - 0xDC00);
/*
        int k1 = k - 0x10000;
        int k2left = k1/0x400;       // get left part
        k2left += 0xD800;
        int k2right = k1%0x400;      // get right part
        k2right += 0xDC00;

        String s2left = TextFormatter.toHexString(k2left);
        String s2right = TextFormatter.toHexString(k2right);

        int m = 0x10000 + k2left * 0x400 + k2right;
        boolean success = k == m;

        String s = TextFormatter.toHexString(n);
 */
    }

    // transforms "\\u0915" to &#x0905;
	static public String slashUToEntity(String s) throws Exception
	{
		s = slashUToChar(s);
		return charToEntity(s);
	}

	static public String entityToSlashU(String s) throws Exception
	{
		s = entityToChar(s);
		return charToSlashU(s);
	}

    static public String intToString(int n, int len)
    {
        String result = Integer.toString(n);
        if (result.length() < len)
        {
            int pad = len - result.length();
            result = "000000000000000000000000000000000".substring(0, pad) + result;
        }
        return result;
    }






    private static String spaces = "                                                                                                                                                     ";
    public static String toCol(StringBuilder sb, int col)
    {
        int margin = col - sb.length();
        if (margin <= 0)
            return "";
        else
            return spaces.substring(0, col - sb.length());
    }

}

