package org.sanskritlibrary.webservice;

import org.sanskritlibrary.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Hashtable;
import java.io.InputStream;

public class TransformMap implements Iterable<TransformMapEntry>
{
	private boolean entireString = false;
	ArrayList<TransformMapEntry> tm = new ArrayList<TransformMapEntry>();
	private String regex = null;
    private String virtualFontname = null;     // Used for converting RTF files. If first FSM has no sharedState attribute then this is fontname used in RTF file, otherwise it is the shared state name.
    private FSM.OptionsUnknown optionUnknown = FSM.OptionsUnknown.FLAG;
    private FSM.OptionsFlagged optionFlagged = FSM.OptionsFlagged.ASIS_NO_DELIMS;
    private char leftFlag = '[';
    private char rightFlag = ']';
    private boolean removeTags;
    private Hashtable<Character, String> mappings = null;

	public TransformMap(String maps, String regex, ResourceInputStream ris,
			  Hashtable<String, FSM[]> fsmh) throws Exception
	{
		this(maps, ris, fsmh);
		this.regex = regex;
	}

	public TransformMap(String maps, ResourceInputStream ris,
			  Hashtable<String, FSM[]> fsmh) throws Exception
	{
		maps = Regex.replace(maps, "\\s+", "");
		if (!maps.equals("*"))
		{
			String[] map = maps.split(";");

			for (String mp : map)
			{
				Regex re = new Regex(mp, "^(.+):(.+)->(.+?)(?:\\[(.*)\\])?$");
				if (re.find())
				{
					boolean retain = false;
					String name = re.getGroup(1);
					if (name.startsWith("<!"))          // should tag be retained in output also
					{
						retain = true;                  // yes
						name = "<" + name.substring(2);
					}
					else if (name.equals("*"))
					{
						if (map.length != 1)
							throw new Exception("If * used there can be only one field in the map");
						entireString = true;
					}
					
					String source = re.getGroup(2);
					String target = re.getGroup(3);
                    String paramString = re.getGroup(4);
                    Hashtable<String, String> params = null;
                    if (paramString != null)
                    {
                        String[] ss = paramString.split("\\s*,\\s*");
                        params = new Hashtable<String, String>();
                        for (String s : ss)
                        {
                            String[] fields = s.split("\\s*:\\s*");
                            if (fields.length == 2)
                                params.put(fields[0], fields[1]);
                            else
                                params.put(fields[0], "?");
                        }
                    }
					tm.add(new TransformMapEntry(name, retain, source, target, params, ris, fsmh));
				}
				else
				{
					throw new Exception("Syntax error in map: " + maps + " (use either '<tagname>:source->target;' or '*:source->target;)");
				}
			}
		}
	}

    public void setOptionUnknown(FSM.OptionsUnknown option) {
        this.optionUnknown = option;
    }

    public FSM.OptionsUnknown getOptionUnknown () {
        return this.optionUnknown;
    }

    public void setOptionFlagged(FSM.OptionsFlagged option) {
        this.optionFlagged = option;
    }

    public FSM.OptionsFlagged getUnknownOption () {
        return this.optionFlagged;
    }

    public void setLeftFlag (char ch) {
        this.leftFlag = ch;
    }

    public char getLeftFlag () {
        return this.leftFlag;
    }

    public void setRemoveTags (boolean b) {
        this.removeTags = b;
    }

    public boolean getRemoveTags () {
        return this.removeTags;
    }

    public void setRightFlag (char ch) {
        this.rightFlag = ch;
    }

    public char getRightFlag () {
        return this.rightFlag;
    }

    public void setRemoveTag(boolean removeTags) {
        this.removeTags = removeTags;
    }

    public void setMappings(Hashtable<Character, String> mappings) {
        this.mappings = mappings;
    }

    public String getMapping(char ch) {
        String target = null;
        if (mappings != null) {
            target = mappings.get(ch);
        }

        return target;
    }

    public FSM getInitialFSM() throws Exception
    {
        if (tm.size() > 1)
            throw new Exception("getInitialFSM is allowable only when *: is used");

        return tm.get(0).fsm[0];
    }

	public String getPattern()
	{
		StringBuilder pattern = new StringBuilder();
		if (tm.size() > 0)
		{
			String delim = "";
			for (TransformMapEntry tme : tm)
			{
				pattern.append(String.format("%s%s", delim, tme.tag));
				delim = "|";
			}
		}
		else
		{
			pattern.append(".*");
		}
		
		return pattern.toString();
	}

    public void setClearState(boolean b) throws Exception
    {
        if (tm.size() > 1)
            throw new Exception("clear state is only meant to be used when RTF is being converted.");

        for (TransformMapEntry tme : tm)
        {
            for (int i=0; i<tme.fsm.length; i++)
            {
                if (tme.fsm[i] != null)
                    tme.fsm[i].setClearState(b);
            }
        }
    }

    public String dumpText(String text)
    {
        StringBuilder sb = new StringBuilder();
        FSM fsm = tm.get(0).fsm[0];
        for (int i=0; i<text.length(); i++)
        {
            char ch = text.charAt(i);
            for (int j=0; j<fsm.getNumEntries(); j++)
            {
                FSMEntry fse = fsm.getEntry(j);
                if (fse.edge.charAt(0) == ch)
                {
                    String s = String.format("%s: %s\n",
                            TextFormatter.toHexString((int)fse.edge.charAt(0), 2),
                            fse.target);
                    sb.append(s);
                    break;
                }
            }
        }

        return sb.toString();
    }

    public void setVirtualFontname(String fontname) throws Exception
    {
        String ssn = getInitialFSM().sharedStateName;
        if (ssn != null)
            virtualFontname = ssn;
        else
            virtualFontname = fontname;
    }

    public String getVirtualFontname()
    {
        return virtualFontname;
    }

    public boolean isDependantCharacter(char ch) throws Exception
    {
        if (tm.size() > 1)
            throw new Exception("clear state is only meant to be used when RTF is being converted.");

        return tm.get(0).fsm[0].isDependantCharacter(ch);       // is it dependendant in the source encoding? Don't care about target encoding (if different than SLP1)
    }


    public boolean isOverriddenCharacter(String key) throws Exception
    {
        if (tm.size() > 1)
            throw new Exception("clear state is only meant to be used when RTF is being converted.");

        return tm.get(0).fsm[0].isOverriddenCharacter(key);       // is it dependendant in the source encoding? Don't care about target encoding (if different than SLP1)
    }

    public String getRegex()
	{
		return regex;
	}

	public Iterator<TransformMapEntry> iterator()
	{
		return tm.iterator();
	}

	public Hashtable<String, InputStream> initHashtable()
	{
		Hashtable<String, InputStream> ish = new Hashtable<String, InputStream>();
		for(TransformMapEntry tme : tm)
		{
			ish.put(tme.source, null);
			ish.put(tme.target, null);
		}

		return ish;
	}

	public boolean isEntireString()
	{
		return entireString;
	}
}
